/*
 * hashicmp.c - count ip src icmp errors with mention of errors sent back
 * 	to a specific ip src
 */

/*
 * Copyright (c) 2005 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>
#include <strings.h>

#include "ourmon.h"
#include "filter.h"
#include "trigger.h"
#include "barthash.c"

static void  __inline
storePort(struct icmpnode *n, int dport);

static void __inline
storeSPort(struct icmpnode *n, int sport);
static void __inline
sizeCheck(struct icmpnode *n, int udplength);

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

#ifdef TESTMAIN
static char *intoa(unsigned int addr);
#endif

static
char spimcode[8]={
0xf8, 0x91, 0x7b, 0x5a, 0x00, 0xff, 0xd0, 0x11,
};

/* called before list can be used 
 */
void init_icmplist (struct icmplist *l) {
	l->entries = 0;
	bzero (&l->hashlist[0], ICMPHSIZE * sizeof(struct icmpnode *));
}

/*
 * search_srcicmp - hash list lookup
 *
 * If found we increment the counters and the total.  
 *	errors are mutually exclusive.
 *
 *	inputs:
 *		l - a list pointer
 *		ipsrc as logical key
 *	returns:
 *		NULL - didn't find it
 *		struct icmpnode *n 
 */ 
struct icmpnode *
search_srcicmp(struct t_entry *te, struct icmplist *l, unsigned int ipsrc, 
	int tcpcount, 
	int udpcount,           /* it's a udp packet sent from src ip*/ 
	int icmpcount,
	int unreach, 
	int redirect, 
	int ttl, 
	int udprcv, 		/* pkt is udp packet sent back to src */
	unsigned int udpport, int honeyflag)
{
	struct icmpnode *n;
	int mod;
	int result;

	/* if no entries, we didn't find it did we
	*/
	if (l->entries == 0) 
		return (NULL);

	/* search the hash/chain list
	*/
	n = l->hashlist[mod=(barthash(ipsrc) % ICMPHSIZE)];
	for (; n!= NULL ; n=n->next) {
		if (n && ( n->ipsrc == ipsrc)) {

			if (honeyflag) {
				n->udpflags |= HONEYNET;
				te->flags |= HONEYNET;
			}
			if (te->darkflag) {
				n->udpflags |= DARKNET;
				te->flags |= DARKNET;
			}
			if (tcpcount) {
				n->tcpcount++;
			}
			else if (udpcount) {
				/* check for one hardwired pattern which is evil spim
				*/
				if (te->psize > 32) {
					if (te->p[0] == 0x04 && te->p[1] == 0) {
						if (bcmp((unsigned char *)(te->p+24), 
						    	spimcode, 8)==0) {
							n->apptags.lower = sVALUE;
						te->appresult.lower = sVALUE;
						}
					}
				}
				/* if udp packet, and we have a payload and
				 * we have global patterns
				 */
				if ((te->psize) &&  (patterns != NULL)) {
					result=patmatch_matchWithHistory(
					te->proto,
					&n->apptags,
					te->p,
					patterns,
					te->psize,
					&te->attempted,
					&te->appresult,
					n->udpcount,
					n->udp_bytes);
					if (result==1) {
						syslog(LOG_ERR,
                                        "error in pattern match of payload");
                                	}
				}
				n->udp_bytes += te->psize;	

				if (udpport != -1) {
					storePort(n, udpport);
				}
				storeSPort(n, te->sport);
				n->udpcount++;
				sizeCheck(n, te->psize);
			}
			else if (icmpcount) {
				n->icmpcount++;
				n->totalicmp++;
			}
			/* inc counts
			*/
			else if (unreach)  {
				n->unreach++;
				n->totalicmp++;
			}
			else if (redirect) {
				n->redirect++;
				n->totalicmp++;
			}
			else if (ttl) {
				n->ttl++;
				n->totalicmp++;
			}
			else if (udprcv) {
				n->udprcv++;
				n->udp_rbytes += te->psize;
			}
			return(n);
		}
	}
	return (NULL);
}

/*
 *      insert_srcicmp
 *
 *      search_srcicmp succeeds or fails. if it succeeds
 *      we don't call insert_srcicmp.  if it fails, we
 *      call this function to create a new hash list
 *      entry and to set counter values.
 *
 *      returns:
 *              1 - ok
 *              0 - malloc failure
 */

int
insert_srcicmp (struct t_entry *te, struct icmplist *l, unsigned int ipsrc, int tcpcount, 
int udpcount, int icmpcount, unsigned int udpport, int honeyflag)
{ 
	struct icmpnode *n;
	struct icmpnode *h, *prev;
	unsigned int mod;
	int i;
	int result;

	n = (struct icmpnode *) malloc(sizeof(struct icmpnode));
	if (n == NULL) {
		syslog(LOG_ERR, "malloc failure in insert_srcudp, entry/count %d",
			l->entries);
		return (0);
	}
	n->totalicmp = n->udpww = n->unreach = n->redirect = n->ttl = 
		n->tcpcount = n->udpcount = n->icmpcount = 0L;
	for (i = 0; i < 6; i++)
		n->sizes[i] = 0; 
	
	n->pcount = 0;
	n->spcount = 0;
	n->udprcv = 0L;
	n->udp_bytes = 0L;
	n->udp_rbytes = 0L;
	n->apptags.lower = n->apptags.upper = 0L;
	n->udpflags = 0;
	if (honeyflag) {
		n->udpflags |= HONEYNET;
		te->flags |= HONEYNET;
	}
	if (te->darkflag) {
		n->udpflags |= DARKNET;
		te->flags |= DARKNET;
	}
        for ( i = 0; i < MAXUDPPORTS; i++) {
                n->dports[i] = -1;
                n->dpcounts[i] = 0;
        }
        for ( i = 0; i < MAXUDPPORTS; i++) {
                n->sports[i] = -1;
                n->spcounts[i] = 0;
        }

	l->entries += 1;
	n->ipsrc = ipsrc;
	n->ipdst = te->dst.s_addr;  
	
	if (tcpcount)
		n->tcpcount = 1;
	else if (udpcount) {
		/* if udp packet, and we have a payload and
		 * we have global patterns
		 */
		if ((te->psize) &&  (patterns != NULL)) {
			result=patmatch_matchWithHistory(
			te->proto,
			&n->apptags,
			te->p,
			patterns,
			te->psize,
			&te->attempted,
			&te->appresult,
			n->udpcount,
			n->udp_bytes);
			if (result==1) {
				syslog(LOG_ERR,
                                "error in pattern match of payload");
                       	}
		}
		n->udp_bytes += te->psize;	
		n->udpcount = 1;
		sizeCheck(n, te->psize);
		if (udpport != -1) {
			storePort(n, udpport);
		}
		storeSPort(n, te->sport);
	}
	else  {
		n->icmpcount = 1;
		n->totalicmp++;
	}

	/* insert the icmpnode in the hash list
	*/
	n->next = NULL;
	h = l->hashlist[mod=(barthash(ipsrc) % ICMPHSIZE)];
	if (h == NULL) {
		l->hashlist[mod] = n;
	}
	else {
		n->next = h;
		l->hashlist[mod] = n;
	}
	return(1);
}

/* udp payload sizes:
 *	40 90 200 500 1000 bigger
 */
static void __inline
sizeCheck(struct icmpnode *n, int udplength)
{
	
	if (udplength <= 40) {
		n->sizes[0]++;
	}
	else if (udplength <= 90) {
		n->sizes[1]++;
	}
	else if (udplength <= 200) {
		n->sizes[2]++;
	}
	else if (udplength <= 500) {
		n->sizes[3]++;
	}
	else if (udplength <= 1000) {
		n->sizes[4]++;
	}
	else {
		n->sizes[5]++;
	}
}

static void __inline
storePort(struct icmpnode *n, int dport)
{
        int i;

        for ( i = 0; i < MAXUDPPORTS; i++) {
                /* have not seen the port and we have room,
                 * therefore store it and increment counter and return
                 */
                if (n->dports[i] == -1) {
                        n->dports[i] = dport;
                        n->dpcounts[i] = 1;
                        n->pcount++;
/*
printf("XXX: %d %d %d\n",
n->dports[i], n->dpcounts[i], n->pcount);
*/
                        return;
                }
                /* found it, increment count, and return
                */
                if (n->dports[i] == dport) {
                        n->dpcounts[i]++;
                        return;
                }
        }
        /* didn't find it, must not be room,
         * too bad
         */
        return;
}

static void __inline
storeSPort(struct icmpnode *n, int sport)
{
        int i;

        for ( i = 0; i < MAXUDPPORTS; i++) {
                /* have not seen the port and we have room,
                 * therefore store it and increment counter and return
                 */
                if (n->sports[i] == -1) {
                        n->sports[i] = sport;
                        n->spcounts[i] = 1;
                        n->spcount++;
/*
printf("XXX: %d %d %d\n",
n->dports[i], n->dpcounts[i], n->pcount);
*/
                        return;
                }
                /* found it, increment count, and return
                */
                if (n->sports[i] == sport) {
                        n->spcounts[i]++;
                        return;
                }
        }
        /* didn't find it, must not be room,
         * too bad
         */
        return;
}

/*
 * deallocate malloced hash entries
 */

void
destroy_icmplist(struct icmplist *l) 
{
	int i;
	struct icmpnode *n;
	struct icmpnode *prev;
	for (i =0; i< ICMPHSIZE; i++) {
		n = l->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->next;
			free(prev);
		}
	}
}

/* pack the hashlist into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct icmpnode *
pack_array(struct icmplist *l)
{
	struct icmpnode *head, *n;
	unsigned int c = l->entries;  /* count of hash entries allocated */
	int i = 0;
	int hcounter = 0;
	int z;

	head = (struct icmpnode *) malloc(c * sizeof (struct icmpnode));
	if (head == NULL) {
		syslog(LOG_ERR, "malloc failure in icmpnode/pack_array, count %d",
			l->entries);
		return (NULL);
	}
	/* walk the hashlist
	*/
	for (i =0 ; i<ICMPHSIZE; i++) {
		n = l->hashlist[i];
		/* if n is non-nulln use, therefore put it in the array,
		* and increment hcounter
		*/
		while (n != NULL) {
			head[hcounter].totalicmp = n->totalicmp;
			head[hcounter].ipsrc = n->ipsrc;
			head[hcounter].ipdst = n->ipdst;
			head[hcounter].tcpcount = n->tcpcount;
			head[hcounter].udpcount = n->udpcount;
			head[hcounter].icmpcount = n->icmpcount;  /* pings */
			head[hcounter].unreach = n->unreach;
			head[hcounter].redirect = n->redirect;
			head[hcounter].ttl = n->ttl;
			head[hcounter].udprcv = n->udprcv;
			head[hcounter].udpflags = n->udpflags;
			head[hcounter].apptags = n->apptags;
			head[hcounter].sizes[0] = n->sizes[0];
			head[hcounter].sizes[1] = n->sizes[1];
			head[hcounter].sizes[2] = n->sizes[2];
			head[hcounter].sizes[3] = n->sizes[3];
			head[hcounter].sizes[4] = n->sizes[4];
			head[hcounter].sizes[5] = n->sizes[5];
			head[hcounter].pcount = n->pcount;
			head[hcounter].spcount = n->spcount;
			head[hcounter].udp_bytes = n->udp_bytes;
			head[hcounter].udp_rbytes = n->udp_rbytes;
/*
printf("XXX: pack: pcount %d\n", head[hcounter].pcount);
*/

                        for (z = 0; z < n->pcount; z++) {
                                head[hcounter].dports[z] = n->dports[z];
                                head[hcounter].dpcounts[z] = n->dpcounts[z];
/*
printf("XXX: %d %d\n", head[hcounter].dports[z], head[hcounter].dpcounts[z]);
*/
                        }
                        for (z = 0; z < n->spcount; z++) {
                                head[hcounter].sports[z] = n->sports[z];
                                head[hcounter].spcounts[z] = n->spcounts[z];
                        }

			n = n->next;
			hcounter++;
			
		}
	}  
	return(head);
}

static 
int
sort_by_totalicmp(e1,e2)
struct icmpnode *e1, *e2;
{
	if (e1->totalicmp > e2->totalicmp) return -1;
	if (e1->totalicmp < e2->totalicmp) return 1;
	return 0;
} 

/* given a icmplist, sort it, and print it out (count entries max)
 * 
 * note: there are two lists here.  
 *	1. for ip src, sorted by max/total icmp errors
 *	2. for ip src, implicit udp list, sorted by Dave Burns weight notion:
 *		|udp send counts - udp recv counts| * icmp errors.
 *
 * inputs:
 *      fd - file for print output
 *      l - hash list
 *      count - max desired outputs for all filters
 */

void
print_icmplist(FILE *fd, struct icmplist *l, int count, int sflag, 
		struct slist *isl, struct slist *usl)
{
	struct icmpnode *head = NULL;
	int i;
	int z;
	int c = l->entries;  /* no of hash elements */
	int hw = 0;
	char bpfexpr[1024];
	/* stormworm hack variables */
	int eviltrigger = -1;	/* no evil trigger yet */
	int evilport = 0;
	int evilport11275 = 0;
	int evilport16275 = 0;
	int evilcount11275;
	int evilcount16275;
	int evilpercent = 0; 
	int totalcount;
	/* end stormworm hack variables */

	/* deal with zero case
	*/
	if (c == 0) {
		fprintf(fd, "icmperror_list: %u : ", c); 
		fprintf(fd, "udperror_list: %u : ", c); 
		goto trigger_check;
	} 

	/* form the hashlist into a packed array
	*/
	head = pack_array(l);
	if (head == NULL) {
		/*
		fprintf(fd, "malloc error in print_tcpicmplist\n");
		*/
		return;
	}
	/* quicksort it for icmp first
	*/
	qsort((void *)head, c, sizeof(struct icmpnode), sort_by_totalicmp);
 
	/* print it out
	*/
	/*
	fprintf(fd, "#ip:totalerror:tcp:udp:icmp:unreach:redirect:ttl\n");
	*/
	fprintf(fd, "icmperror_list: %u : ", l->entries); 
	for (i = 0; i < c; i++) {

			
		if ( i >= count) {
			break;
		}
		/*
		 * format: ip address: total icmp errors: tcp: udp :icmp pkts: icmp unreach : redirect: 
		 *       ttl : udpflags :
		 */
		fprintf(fd, "%s:%u:%u:%u:%u:%u:%u:%u:%x:",
			intoa(head[i].ipsrc), 
			head[i].totalicmp, 		/* count of icmp errors */
			head[i].tcpcount, 
			head[i].udpcount, 
			head[i].icmpcount, 
			head[i].unreach,
			head[i].redirect,
			head[i].ttl,
			head[i].udpflags);
	}
	fprintf(fd, "\n");

	/* compute the weights for the udp specific view. 
 	 * overload totalicmp to use it as the weight
	 * the general idea: weight is udp packets sent * icmperrors + rcv count.
	 *
 	 * if busy you appear in the list:
	 * if scanner you appear high
	*/
	for (i = 0; i < c; i++) {
		unsigned int icmperrs = head[i].unreach + 
			head[i].redirect + head[i].ttl;

		if (icmperrs == 0)
			icmperrs = 1;
		head[i].totalicmp = 
			(head[i].udpcount * icmperrs) + head[i].udprcv; 

#ifdef NOPE
		/* if more udp recv then sent, weight is 
		 * send count * icmp errors + recv count.
		 * This way the topN by weight rank will at least
		 * include p2p hosts who recv more sending pkts,
		 * but will still tend to push scanners to the top.
		*/
		if (head[i].udpcount <= head[i].udprcv) {
		} 

		/* else compute weight for more send than recv
		*/
		else {
			head[i].totalicmp = (head[i].udpcount - head[i].udprcv) * icmperrs;
		}
#endif
	}

	/* quicksort it again.  note name overload.
	*/
	qsort((void *)head, c, sizeof(struct icmpnode), sort_by_totalicmp);

	/* now print out udperror_list
	*/
	fprintf(fd, "udperror_list: %u : ", l->entries); 
	for (i = 0; i < c; i++) {

		if ( i >= count) {
			break;
		}

		/* look up l3 and l4 info if scanflag is set
		*/
		if (sflag) {
			head[i].l3count = scanLookup(isl, head[i].ipsrc);
			head[i].l4count = scanLookup(usl, head[i].ipsrc);
		}
		else {
			head[i].l3count = head[i].l4count = 0;
		}

		/*
		 * format: ip address: weight: udp pkts sent, udp pkts recv: total icmp errors, pings,
		 *	l3 count, l4 count, udpflags, apptags.lower, apptags.upper, sizes(4), dst port count
		 */
		fprintf(fd, "%s:%u:%u:%u:%u:%u:%u:%u:%x:%x:%x:%u:%u:%u:%u:%x:%x:%u:%u:%u:%u:",
			intoa(head[i].ipsrc),   /* 0 */ 
			head[i].totalicmp, 	/* 1 weight */
			head[i].udpcount, 	/* 2 udp sent pkts */
			head[i].udprcv,		/* 3 udp recv pkts */
			head[i].unreach, 	/* 4 udp unreach errors */
			head[i].icmpcount, 	/* 5 pings sent by src */
			head[i].l3count, 	/* 6 unique L3 destinations */
			head[i].l4count, 	/* 7 unique L4 destinations */
			head[i].udpflags, 	/* 8 hardwired app flags */
			head[i].apptags.lower, 	/* 9 PCRE app flags */
			head[i].apptags.upper, 	/* 10 PCRE app flags */
			head[i].sizes[0], 	/* 11 size */
			head[i].sizes[1], 	/* 12 size */
			head[i].sizes[2], 	/* 13 size */
			head[i].sizes[3], 	/* 14 size */
			head[i].sizes[4], 	/* 15 size */
			head[i].sizes[5], 	/* 16 size */
/* new attributes */
			head[i].udp_bytes,      /* 17 sent bytes */
			head[i].udp_rbytes,     /* 18 recv bytes */
			head[i].pcount,		/* 19 udp dst port tuple count */
			head[i].spcount 	/* 20 udp src port tuple count */
			);


/*
printf("XXX print: %d\n", head[i].pcount);
*/
		/* port sigs */
		for (z = 0; z < head[i].pcount; z++) {
			fprintf(fd, "%u,%u,", head[i].dports[z], head[i].dpcounts[z]);
		}
		fprintf(fd, ":");
	}
	fprintf(fd, "\n");
	
	/* pre check trigger condition.  this function tells us 
	 * if the trigger WOULD go off but does not change any
	 * trigger state.  It is more/less boolean.  if true,  
	 * a dynamic bpf filter is set up here to capture all packets 
	 * for statically configured packet counts 
	 * based on the topn udp_error host address
	 * syntax:  src IPADDR and udp
	 */
	hw  = head[0].totalicmp;	/* note: this is UDP work weight */

        if(pre_trigger_check(TOPNUDPERR, hw)) {
                char bpfexpr[1024];
                char *e=NULL;

		sprintf(bpfexpr, "src %s and udp",
                                intoa(head[0].ipsrc));
		e = trigger_dynbpf_compile(TOPNUDPERR, bpfexpr);
                if( e != NULL ) {
			fprintf(fd, "elog: topn_udp_error bpf dynamic parse failure: %s\n", e);
                }
        }

trigger_check:
	/* now actually check the TOPNUDPERROR trigger.
	 * hw is the ww count for the top udp error IP src if any
	 * note: we come here in the zero list case because we might
	 * be turning the trigger off
	 */
	trigger_condition_check(TOPNUDPERR, fd, hw, "topn_udp_err");

	free(head);
}          

#ifdef TESTMAIN

static int f = 0;

unsigned int
getRandomWord()
{
	static unsigned int rw;

#define REALLYRANDOM
#ifdef REALLYRANDOM
	if (f == 0) {
		f = open("/dev/urandom", 0);
		if (f < 0) {
			perror("open");
			exit(1);
		}
	}
	read(f, &rw, sizeof(unsigned int));
#else
	rw = rand();
#endif
	return(rw);
}

functionTest()
{
	struct icmplist l;
	unsigned int ipsrc;
	int rc;
	struct icmpnode *n = NULL;

	/* load sample ipkey and counter
	*/
	ipsrc = inet_addr("1.1.1.1");

	init_icmplist(&l);

	n = search_srcudp(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("search failed\n");
	} 
	
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	n = search_srcudp(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 

	/* add 4 more nodes and dump the list
	*/
	ipsrc = inet_addr("1.1.1.2");
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.3");
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}
	ipsrc = inet_addr("1.1.1.4");
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.5");
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	ipsrc = inet_addr("1.1.1.6");
	rc = insert_srcudp(&l, ipsrc, 1, 0, 0);
	if (rc == 0) {
		printf("insertNode failed\n");
	}

	/* now search for a given node, and increment its
	 * value
	 */
	n = search_srcudp(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	/* now search for a previous node, note last
	 * as a simple cache should not be used here
	 */
	ipsrc = inet_addr("1.1.1.4");
	n = search_srcudp(&l, ipsrc, 1, 0, 0);
	if ( n == NULL ) {
		printf("searchNode failed\n");
	} 
	else {
		printf("searchNode found node\n");
	}

	dump_hash(&l);

	print_icmplist(stdout, &l, 10, NULL, NULL);

	destroy_icmplist(&l);
}


main()
{
	functionTest();
}
#endif



