/*
 * hashsort.c - deal with dynamic allocation and sorting
 *	of an arbitrary number of nodes.  This is intended
 *	for use by the top n mechanism.
 */


/*
 * Copyright (c) 2001, 2006 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <syslog.h>
#include <pcap.h>
#include <string.h>

#include "ourmon.h"
#include "filter.h"
#include "stats.h"

#include "trigger.h"

/* ehash is better at spreading out into buckets
 * given 8 bytes of ip src and ip dst
 */

#ifdef EHASH
#include "ehash.c"	/* inlined function */
#else
/* bart massey's sacred hash
*/
#include "barthash.c"	/* inlined function */
#endif

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

#define USETCP		1
#define USEUDP  	2
#define USEICMP 	3
#define NOTINUSE 	0

extern char *prototostring(int);
static void printFlow(FILE *fd, struct node *head);  
static void printPerl(FILE *fd, struct node *head);  
static void printPerlFormat(FILE *fd);
static void printHex(FILE *fd, struct node *head);  

char *intoa(unsigned int);

/* called before list can be used
*/
void
initList(struct list *l)
{
	l->count = 0;
	l->last = NULL;
	l->total = 0;
	l->tcpcount = l->udpcount = l->icmpcount = 0;

	/* zero out hash array
	*/
	bzero(&l->hashlist[0], HASHSIZE * sizeof(struct node *));
}

static void __inline
sizecalc(struct t_entry *te, struct node *n, struct ipkey *ipkey)
{

	/* set counter if payload
	*/
	/* store size histogram
	*/
	if (ipkey->ipproto == IPPROTO_TCP) {

		if (te->synflag)	
			n->c1++;
		else if (te->finflag)	
			n->c2++;
		else if (te->resetflag)	
			n->c3++;

		/* there exists layer 7 data
		*/
		if (te->psize)
			n->payload++;

		/* if no payload set first size bucket
		 * this is because tcp has options
		*/
		if (!te->psize) {
			n->sizes[0]++;
		}
		else if (te->bytes < 600) {
			n->sizes[1]++;
		}
		else if (te->bytes < 1460) {
			n->sizes[2]++;
		}
		else {
			n->sizes[3]++;
		}
	}
	else if (ipkey->ipproto == IPPROTO_UDP) {

		if (te->psize)
			n->payload++;

		/* udp has no options
		*/
		if (te->bytes == 28) {
			n->sizes[0]++;
		}
		else if (te->bytes < 600) {
			n->sizes[1]++;
		}
		else if (te->bytes < 1000) {
			n->sizes[2]++;
		}
		else {
			n->sizes[3]++;
		}
	}
	else if (ipkey->ipproto == IPPROTO_ICMP) {
		/* these bins are just rough and not
		 * based on anything heuristic-wise
		*/
		if (te->bytes < 100) {
			n->sizes[0]++;
		}
		else if (te->bytes < 600) {
			n->sizes[1]++;
		}
		else if (te->bytes < 1000) {
			n->sizes[2]++;
		}
		else {
			n->sizes[3]++;
		}
	}
	/* need to work on this one - some kinda tunnel
	*/
	else {
		/* these bins are just rough again
		*/
		if (te->bytes < 100) {
			n->sizes[0]++;
		}
		else if (te->bytes < 600) {
			n->sizes[1]++;
		}
		else if (te->bytes < 1000) {
			n->sizes[2]++;
		}
		else {
			n->sizes[3]++;
		}
	}
}


#ifdef BOTMESH
/* given a flow n, and the hashlist l, and an ipkey for flow n,
 * find out if there is a 2-way "equal and opposite" flow,
 * if so, set a pointer in n to point to that flow.
*/
static void __inline
twowayCheck(struct node *n, struct list *l, struct ipkey *ipkey)
{
	int mod2;
	struct node *m2, *n2;
	struct ipkey peer;

	/* lookup the flow in the hash array again in 
	 * inverted order (IP src, L4 src and IP dst, L4 dst
	 * swap places)
	 * this is a 2-way flow aggregation
	*/
	/* load key in reverse order
	*/
	peer.ipdst = ipkey->ipsrc;
	peer.ipsrc = ipkey->ipdst;
	peer.srcport = ipkey->dstport;
	peer.dstport = ipkey->srcport;
	peer.ipproto = ipkey->ipproto;  
	
#ifdef EHASH
	mod2=ehash((unsigned char *)&peer, sizeof(struct ipkey)) % HASHSIZE;
#else
	{
		/* hash lookup of key:
		 * ipsrc, ipdst, dst port src port only at this point
		 */
		unsigned int x,y,z,h;
		x = barthash(peer.ipsrc);
		y = barthash(peer.ipdst);
		z = barthash((peer.dstport << 16) ^ peer.srcport);
		mod2 = (x ^ y ^ z) % HASHSIZE;
	}
#endif

	/* we either have a hash chain/bucket or no */
	m2 = n2 = l->hashlist[mod2];

	/* if peer chain found, find individual node in chain
	 * else ignore
	*/
	if (m2) {
		for (; n2 != NULL; n2 = n2->hnext) {
			/* found it
			*/
			if (n2 && (
				(n2->ipkey.ipdst == ipkey->ipsrc) &&
				(n2->ipkey.ipsrc == ipkey->ipdst) &&
				(n2->ipkey.dstport == ipkey->srcport) &&
				(n2->ipkey.srcport == ipkey->dstport) &&
				(n2->ipkey.ipproto == ipkey->ipproto)))  {
				/* might as well have both point to each
				 * other
				*/

/*
printf("GOOMBAH: %x %x %x %x %x: %x %x %x %x %x\n", 
n2->ipkey.ipdst, n2->ipkey.ipsrc, n2->ipkey.dstport, n2->ipkey.srcport, n2->ipkey.ipproto, 
ipkey->ipsrc, ipkey->ipdst, ipkey->srcport, ipkey->dstport, ipkey->ipproto);
*/

				if (!n->twoway) {
/*
printf("GOOMBAH: %d %d\n", n->ipkey.ipproto, n2->ipkey.ipproto);
*/
					n2->twoway = n;
					n->twoway = n2;
				}
				break;
			}
		}
	}
}
#endif


/*
 * search and insert the node if not found by search
 * count as well...
 */
void
searchInsertNode(
struct t_entry *te, struct list *l, struct ipkey *ipkey, COUNTER counter, int flowcap) 
{
	struct node *n,*n2;
	struct node *m,*m2;
	struct node *h;
	int first = 0;
	int mod;
	int mod2;
	struct node *prevnode;

	/* search the hash/chain list 
	*/
	/* SCATTER */
#ifdef EHASH
buggy
	mod=ehash((unsigned char *)ipkey, sizeof(struct ipkey)) % HASHSIZE;
	m = n = l->hashlist[mod];
#else
	{
		/* hash lookup of key:
		 * ipsrc, ipdst, dst port src port only at this point
		 */
		unsigned int x,y,z,h;
		x = barthash(ipkey->ipsrc);
		y = barthash(ipkey->ipdst);
		z = barthash((ipkey->dstport << 16) ^ ipkey->srcport);
		mod = (x ^ y ^ z) % HASHSIZE;
		m = n = l->hashlist[ mod ];
	}
#endif

	/* walk buckets
	*/
	for (; n != NULL; n = n->hnext) {
		/* we found a match
		*/
		if (n && (
			(n->ipkey.ipdst == ipkey->ipdst) &&
			(n->ipkey.ipsrc == ipkey->ipsrc) &&
			(n->ipkey.dstport == ipkey->dstport) &&
			(n->ipkey.srcport == ipkey->srcport) &&
			(n->ipkey.ipproto == ipkey->ipproto) 
			)) {
#ifdef BOTMESH
			/* note: if already two way, no need to check again
			*/
			if (flowcap && !n->twoway)
				twowayCheck(n, l, ipkey);
#endif
			sizecalc(te, n, ipkey); 
			n->counter += counter;
			n->pktcount++;
			/* copy tags if checked, 
			 * basic idea: the packet has several tags fields.
			 * we OR them into our per flow data structure if non-zero.
			 * we won't bother to do this
			 * on insert.  we are statistical after all :-> or :-<
			*/
			if (te->flags) {
				n->flags |= te->flags;
			}
			if (te->appresult.upper) {
				n->apptags.upper |= te->appresult.upper;
			}
			if (te->appresult.lower) {
				n->apptags.lower |= te->appresult.lower;
			}
			l->total += counter;
			if (first == 0) {
				os.hs_bfirst++;
			}
			else {
				os.hs_bother++;
			}
			return;	/* search succeeded */
		}
		first++;
	}

	/* commit to insert
	*/
	os.hs_insert++;

	n = (struct node *) malloc(sizeof(struct node)); 

	/* out of space. ouch.
	*/
	if (n == NULL) {
		syslog(LOG_ERR, "topn: malloc out of space, count %d\n", l->count);
		return;
	}

	l->count++; 
	if (ipkey->ipproto == IPPROTO_TCP)
		l->tcpcount++;
	else if (ipkey->ipproto == IPPROTO_UDP)
		l->udpcount++;
	else if (ipkey->ipproto == IPPROTO_ICMP)
		l->icmpcount++;
	n->ipkey = *ipkey; 
	n->counter = counter;
	n->pktcount = 1;
	n->flags = n->apptags.upper = n->apptags.lower = 0L;
	n->payload = 0;
#ifdef BOTMESH
	if (flowcap) {
		n->twoway = NULL;
		twowayCheck(n, l, ipkey);
	}
#endif
	n->c1 = n->c2 = n->c3 = 0;
	n->sizes[0] = n->sizes[1] = n->sizes[2] = n->sizes[3] = 0;

	/* set payload/size buckets
	*/
	sizecalc(te, n, ipkey); 

	l->total += counter;

	/* insert the node in the hash list
	*/

	n->hnext = NULL;
		
	if (m == NULL) {
		l->hashlist[mod] = n;
	}
	/* collision, put new node in 1st
	*/
	else {
		os.hs_coll++;
		n->hnext = m;
		l->hashlist[mod] = n;
	}
	return;
}

void
debugList(struct list *l)
{
	dumpHash(l);
}

/*
 * deallocate malloced hash entries
 */
void
destroyList(struct list *l) 
{
	int i;
	struct node *n;
	struct node *prev;

	for ( i = 0; i < HASHSIZE; i++) {
		n = l->hashlist[i];
		while(n != NULL) {
			prev = n;
			n = n->hnext;
			free(prev);
		}
	}
}

dumpHash(struct list *l)
{
	struct node *h;
	int counters[HASHSIZE];
	int i;
	int ls = 0;

	printf("count %d, total %d\n", l->count, l->total);
	bzero(&counters, sizeof(int) * HASHSIZE);

	printf("\nhash dump--------------\nlist count %d\n", l->count);
	for (i = 0; i < HASHSIZE; i++) {
		h = l->hashlist[i];
		while(h != NULL) {
			ls = 1;
			printf("[%d]: ", i);
			printf("ipsrc %s<->", intoa(h->ipkey.ipsrc));
			printf("ipdst %s ", intoa(h->ipkey.ipdst));
			printf("count %d :", h->counter);
			counters[i]++;  /* total buckets in tier used */
			h = h->hnext;
		}
		if (ls) {
			printf(":[%d]\n", i);
		}
		ls = 0;
	}
	printf("\n");
	for (i = 0; i < HASHSIZE; i++) {
		if (counters[i] != 0) {
			printf("counter[%d]: %d buckets\n", i, counters[i]);
		}
	}
	printf("end hash dump--------------\n");
}

/*
 * ok, we have a hash list.  pack it into an array
 * and return the array.
 *
 *	returns:
 *		1 - packed array exists
 *		0 - malloc failure
 */
static
struct node *
packArray(struct list *l)
{
	struct node *head, *n;
	unsigned int c = l->count;	/* count of hash entries allocated */
	int i = 0;
	int hcounter = 0;

	/* create the "array"
	*/
	head = (struct node *) malloc(c * sizeof (struct node));
	if (head == NULL) {
		syslog(LOG_ERR, "topn: packArray out of space, count %d\n",
			l->count);
		return(NULL);
	}

	/* walk the hashlist
	*/
	for ( i = 0; i < HASHSIZE; i++) {
		n = l->hashlist[i];
		/* if n is non-null, we have a node
		 * in use, therefore put it in the array,
		 * and increment hcounter
		 */
		if (n != NULL) {
			os.hs_ecount++;	/* number of full first buckets */
		}
		while (n != NULL) {
			head[hcounter].ipkey = n->ipkey; 
			head[hcounter].counter = n->counter;
			head[hcounter].pktcount = n->pktcount;
			head[hcounter].flags = n->flags;
			head[hcounter].apptags.upper = n->apptags.upper;
			head[hcounter].apptags.lower = n->apptags.lower;
			head[hcounter].payload = n->payload;
			head[hcounter].sizes[0] = n->sizes[0];
			head[hcounter].sizes[1] = n->sizes[1];
			head[hcounter].sizes[2] = n->sizes[2];
			head[hcounter].sizes[3] = n->sizes[3];
			head[hcounter].c1 = n->c1;
			head[hcounter].c2 = n->c2;
			head[hcounter].c3 = n->c3;
			head[hcounter].twoway = n->twoway;	/* note that this is pointer to peer hashlist node */
			hcounter++;
			/* walk to end of linked list repeating previous
			 * operation
			 */
			n = n->hnext;
		}
	}
	return(head);
}

/* qsort comparison - sort by byte count
*/
static int      
sortbycounter(e1, e2)
struct node *e1, *e2;
{
        if (e1->counter > e2->counter) return -1;
        if (e1->counter < e2->counter) return 1;
        return 0;
}       

/* qsort comparison - sort by pkt count
*/
static int      
sortbypktcount(e1, e2)
struct node *e1, *e2;
{
        if (e1->pktcount > e2->pktcount) return -1;
        if (e1->pktcount < e2->pktcount) return 1;
        return 0;
}       

char *
intoa(addr)
unsigned int addr;
{
	register char *cp;
	register unsigned int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

    	cp++;
	return cp;
}

/*
 * stoopid: but it gives us the ability to do this
 * twice.  for ip src and ip dst.
 *
 * A faster replacement for inet_ntoa().
 */
char *
intoa2(addr)
unsigned int addr;
{
	register char *cp;
	register unsigned int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do {
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if (byte > 0) {
			*--cp = byte % 10 + '0';
			byte /= 10;
			if (byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);

    	cp++;
	return cp;
}

static
int
biggest_value(int tv, int uv, int iv)
{
	int big;
	if (tv > uv)
		big = tv;
	else  
		big = uv;

	if (big < iv)
		big = iv;

	return(big);
}


/*
 * given a list, sort it,
 * and print topN results.
 * if tcp flag, print top N tcp.
 * if udp flag, print top N udp,
 *
 * inputs:
 *	fd - file for print output
 *	l - hash array
 *	anyflag/tcpflag/udpflags - flags to indicates various filter outputs
 *	count - max desired outputs for all filters
 */
printList(FILE *fd, struct list *l, int anyflag, int tcpflag, int udpflag, int icmpflag, int count, int flowcap, char *flowfile, int ssh, char *sshfile) 
{
	struct node *head;
	char *ps;
	char buf[1024];
	char b1[1024], b2[1024];
	int i;
	unsigned int subtotal = 0;
	unsigned int total = l->total;
	int protohasports = 0;
	int tcpcount, udpcount, icmpcount;
	int tv = 0;
	int tvalue = 0;
	static int cur_trigger = NOTINUSE;
	int big;
	FILE *afd;
	FILE *bfd;
	FILE *cfd;
	FILE *sfd;
	FILE *sshfd;
	FILE *zfd;

#ifdef BOTMESH
	int bm_total, bm_totaltcp, bm_totaludp, bm_totalother, 
		bm_nopayload, bm_sizes1, bm_tcpsizes1, bm_udpsizes1, ts;
	int bm_twoway = 0;
	int bm_twowaysmall = 0;
	int bm_c1 = 0;
	int bm_c2 = 0;
#endif

	int c = l->count;	/* no of hash elements */

	tcpcount = udpcount = icmpcount = 0;

	/* form the hashlist into a packed array
	*/
	head = packArray(l);
	if (head == NULL) {
		/*
		fprintf(fd, "malloc error in printList\n");
		*/
		return;
	}

	/* quicksort it
	*/
	qsort((void *)head, c, sizeof (struct node), sortbycounter);

	/* print it out
	*/
	if (anyflag) {
		/* produce topn_ip
		*/
		fprintf(fd, "topn_ip : %u : ", c); 
		/* walk thru all elements of hash list
		*/
		for ( i = 0; i < c; i++) {
			/* if i reaches config count
			 * quit
			 */
			if (i >= count) {
				break;
			}
			/*
			subtotal += head[i].counter;
			*/
			/* produce a string for the proto type
			 * if possible, else a number.
			 */
			ps = prototostring(head[i].ipkey.ipproto);
			if (ps == NULL) {
				sprintf(buf, "%d", head[i].ipkey.ipproto);
				ps = &buf[0];
			}
			else {
				protohasports = 1;
			}
			/* intoa can only do one ip address at a time,
			 * therefore we must buffer one
			 */
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			/* if we have non-zero ports include them,
			 * else drop them.
			 */
			if (protohasports) {
				fprintf(fd,"%s.%d->%s.%d(%s): %u:%x:%x:%x: ",
					intoa(head[i].ipkey.ipsrc),
					head[i].ipkey.srcport,
					b1,
					head[i].ipkey.dstport,
					ps,
					head[i].counter,
					head[i].flags,
					head[i].apptags.lower,
					head[i].apptags.upper 
					);
			}
			else {
				fprintf(fd,"%s->%s(%s):%u:%x:%x:%x: ",
					intoa(head[i].ipkey.ipsrc),
					b1,
					ps,
					head[i].counter,
					head[i].flags,
					head[i].apptags.lower,
					head[i].apptags.upper 
					);
			}
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}
	if (tcpflag) {
		fprintf(fd, "topn_tcp : %u : ", l->tcpcount); 
		for ( i = 0; i < c; i++) {
			if (head[i].ipkey.ipproto != IPPROTO_TCP)
				continue;
			tcpcount++;
			if (tcpcount > count) {
				break;
			}
			/*
			subtotal += head[i].counter;
			*/
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			fprintf(fd, "%s.%d->%s.%d: %u:%x:%x:%x: ",
				intoa(head[i].ipkey.ipsrc),
				head[i].ipkey.srcport,
				b1,
				head[i].ipkey.dstport,
				head[i].counter,
				head[i].flags,
				head[i].apptags.lower,
				head[i].apptags.upper 
				);
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}
	if (udpflag) {
		fprintf(fd, "topn_udp : %u : ", l->udpcount); 
		for ( i = 0; i < c; i++) {
			if (head[i].ipkey.ipproto != IPPROTO_UDP) {
				continue;
			}
			udpcount++;
			if (udpcount > count) {
				break;
			}
			/*
			subtotal += head[i].counter;
			*/
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			fprintf(fd, "%s.%d->%s.%d: %u:%x:%x:%x: ",
				intoa(head[i].ipkey.ipsrc),
				head[i].ipkey.srcport,
				b1,
				head[i].ipkey.dstport,
				head[i].counter,
				head[i].flags,
				head[i].apptags.lower,
				head[i].apptags.upper 
				);
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}
	if (icmpflag) {
		fprintf(fd, "topn_icmp : %u : ", l->icmpcount); 
		for ( i = 0; i < c; i++) {
			if (head[i].ipkey.ipproto != IPPROTO_ICMP)
				continue;
			icmpcount++;
			if (icmpcount > count) {
				break;
			}
			/* ip dst put in string b1 */
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			/* ipsrc.majorcode->ipdst.minorcode 
			 * which needs to be pointed out in the documentation
			*/
			fprintf(fd, "%s.%d->%s.%d: %u:%x:%x:%x: ",
				intoa(head[i].ipkey.ipsrc),
				head[i].ipkey.srcport,
				b1,
				head[i].ipkey.dstport,
				head[i].counter,
				head[i].flags,
				head[i].apptags.lower,
				head[i].apptags.upper 
				);
		}
		/*
		fprintf(fd, " xtra: %u\n", total - subtotal);
		*/
		fprintf(fd, "\n");
	}

	/* quicksort for total pkt cont
	*/
	qsort((void *)head, c, sizeof (struct node), sortbypktcount);

#ifdef BOTMESH
	/*
	fprintf(fd, "topn_ip : %u : ", c); 
	*/
	/* walk thru all elements of hash list
	*/
	bm_total = c;
	bm_totaltcp = l->tcpcount;
	bm_totaludp = l->udpcount;
	bm_totalother = c-(bm_totaltcp + bm_totaludp);
	bm_nopayload = bm_sizes1 = 0;
	bm_tcpsizes1 = bm_udpsizes1 = 0;

#ifdef BOTDEBUG
	/* "only" size1 flows */
	sfd = fopen("../output/bmsizes1.txt", "w");

	/* all of them */
	afd = fopen("../output/allflows.txt", "w");
	/* unknown protocols */
	bfd = fopen("../output/unkflows.txt", "w");
#endif

	if (flowcap) {
		zfd = fopen(flowfile, "w");
	}

	if (flowcap) {
		for ( i = 0; i < c; i++) {
			int bm_p1small = 0;


			/* following is 1-way stats
			*/
#ifdef BOTDEBUG
			/* print all flows (debug) */
			printFlow(afd, head+i);
			/* print non-tcp/udp flows */
			if (head[i].ipkey.ipproto == IPPROTO_TCP)
				;
			else if (head[i].ipkey.ipproto == IPPROTO_UDP)
				;
			else 
				printFlow(bfd, head+i);
#endif
			if (head[i].payload == 0) {
				bm_nopayload++;
			}
			/* FILTER 1 */
			/* see if flow is smallish (non-zero but not large payload)
			 * headers are not ruled out (smallest packets are ok)
			*/
			ts = head[i].sizes[2] + head[i].sizes[3];
			if (head[i].sizes[1] && (ts == 0)) {
				bm_sizes1++;
				bm_p1small = 1;		/* first flow is small */
#ifdef BOTDEBUG
				printFlow(sfd, head+i);
#endif
			}
			/* FILTER 2 */
			/* see if the flow is 2-way and both are small*/
			if (head[i].twoway) {
				struct node *peer;

				bm_twoway++;	/* count number of twoway flows */
				if (head[i].ipkey.ipproto == IPPROTO_TCP)
					bm_c1++;

				peer = head[i].twoway;	
				
				/* see if the flow has been dealt with before as a peer
				*/
				if (peer->twoway) {
					/* if this flow was small, see if the related flow is small
					*/
					if (bm_p1small) {
						ts = peer->sizes[2] + peer->sizes[3];
						/* FILTER 3 */
						/* we still ignore the flow if both addresses are inside
						*/
						if (ishome(head[i].ipkey.ipsrc) &&
						ishome(head[i].ipkey.ipdst)) 
								;
						/* print both */
						else if (peer->sizes[1] && (ts == 0)) {
							/* if any syns or fins count in bm_c1
							*/
							if (head[i].c1 || head[i].c2) {
/*
								bm_c1++;
*/
							}
							bm_twowaysmall++;
							printPerl(zfd, head+i);
							printPerl(zfd, head[i].twoway);
							if (head[i].ipkey.ipproto == IPPROTO_TCP)
								bm_tcpsizes1++;
							else if (head[i].ipkey.ipproto == IPPROTO_UDP)
								bm_udpsizes1++;
						}
					}
					peer->twoway->twoway = NULL;
				}
			}
		}
	}

#ifdef BOTDEBUG
	fclose(sfd);
	fclose(afd);
	fclose(bfd);
#endif
	if (flowcap) {
		fclose(zfd);

		/* mon.lite stats on how this is doing 
		 * 1. total flows:
		 * 2. total tcp flows:
		 * 3. total udp flows:
		 * 4. total other (not tcp or udp) flows:
		 * 5. flows with no payload (only valid if tcp/udp)
		 * 6. flows with size1 bucket set
		 * 7. tcp 2-way flows with size1 bucket set
		 * 8. udp 2-way flows with size1 bucket set
		 * 9. total 2-way flows:
		 * 10. total 2-way flows with size1 bucket set
		 * 11. total syn/fin flows with size1 bucket set
		 * graph:
		 *	total flows, total tcp flows, tcp 2-way flows with size1 bucket * 2, 
		*/
		fprintf(fd, "bot_mesh : %u : %u : %u : %u : %u : %u : %u : %u: %u :%u: %u:\n", 
			bm_total, bm_totaltcp, bm_totaludp, bm_totalother,
			bm_nopayload, bm_sizes1, bm_tcpsizes1,
			bm_udpsizes1, bm_twoway, bm_twowaysmall, bm_c1);  
	}
#endif /* BOTMESH */

	/* print it out
	*/
	if (anyflag) {
		/* produce topn_ip
		*/
		fprintf(fd, "topn_ip_pkts : %u : ", c); 
		/* walk thru all elements of hash list
		*/
		for ( i = 0; i < c; i++) {
			/* if i reaches config count
			 * quit
			 */
			if (i >= count) {
				break;
			}
			/* produce a string for the proto type
			 * if possible, else a number.
			 */
			ps = prototostring(head[i].ipkey.ipproto);
			if (ps == NULL) {
				sprintf(buf, "%d", head[i].ipkey.ipproto);
				ps = &buf[0];
			}
			else {
				protohasports = 1;
			}
			/* intoa can only do one ip address at a time,
			 * therefore we must buffer one
			 */
			sprintf(b1, "%s", intoa(head[i].ipkey.ipdst));
			/* if we have non-zero ports include them,
			 * else drop them.
			 */
			if (protohasports) {
				fprintf(fd,"%s.%d->%s.%d(%s): %u:%x:%x:%x: ",
					intoa(head[i].ipkey.ipsrc),
					head[i].ipkey.srcport,
					b1,
					head[i].ipkey.dstport,
					ps,
					head[i].pktcount,
					head[i].flags,
					head[i].apptags.lower,
					head[i].apptags.upper 
					);
			}
			else {
				fprintf(fd,"%s->%s(%s):%u:%x:%x:%x: ",
					intoa(head[i].ipkey.ipsrc),
					b1,
					ps,
					head[i].pktcount,
					head[i].flags,
					head[i].apptags.lower,
					head[i].apptags.upper 
					);
			}
		}
		fprintf(fd, "\n");
	}

	/* save the ssh data to a file.  tcp only,  port 22 mentioned
	*/
	if (ssh) {
		if (sshfd = fopen(sshfile, "w")) {
			fprintf(sshfd, "# SSH FILE!\n"); 
			printPerlFormat(sshfd);
			for ( i = 0; i < c; i++) {
				if (head[i].ipkey.ipproto == IPPROTO_TCP)
					if (head[i].ipkey.srcport == 22 || head[i].ipkey.dstport == 22) {
						printPerl(sshfd, head + i);
					}
			}
			fclose(sshfd);
		}
		else {
			/* TBD - need syslog error message for these things
			*/
		}
	}

	
	free(head);

	fprintf(fd, "#hs_stat: 0: 0: bfirst: bother: bmiss: coll: insert: fcount:\n");
        fprintf(fd, "hs_stat: %u: %u: %u: %u: %u: %u: %u: %u:\n",
                        os.hs_chits, 	/* dead */
			os.hs_cmiss, 	/* dead */
			os.hs_bfirst,
                        os.hs_bother, 
			os.hs_bmiss, 	/* dead */
			os.hs_coll,
                        os.hs_insert, 
			os.hs_ecount  
			);

	/* if the trigger is configured on, and the current value is
	 * greater than the threshold value AND the state of the trigger is off,
	 * dynamically arm the trigger.
	 *
	 *	l->tcpcount, l->udpcount, l->icmpcount.
	 * 	If any of these is greater than the value, then
	 *	set a dynamic bpf.  do this as follows:
	 *	0 to 3 of them may be greater than the value.
	 *	if only 1, set bpf to that one.
	 *	if more than 1, make bpf "ip".
	*/
	big = biggest_value(l->tcpcount, l->udpcount, l->icmpcount);

	tv = pre_trigger_check(TOPNIP, big);
	if(tv) {
                char fname[1024];
		char *e = NULL;
		char *bpfexpr;
		int totalcount = l->count/30;		/* convert to flows per sec */
		int tcount = l->tcpcount/30;
		int ucount = l->udpcount/30;
		int icount = l->icmpcount/30;

		/* first figure out which bpf expression to use
		 * can we narrow it?
		*/
		bpfexpr = "ip";		/* this is the default */
		if ((icount > tv) && ((tcount < tv) && (ucount < tv))) {
			bpfexpr = "icmp";
			cur_trigger = USEICMP;	
		}
		else if ((ucount > tv) && ((tcount < tv) && (icount < tv))) {
			bpfexpr = "udp";
			cur_trigger = USEUDP;	
		}
		else if ((tcount > tv) && ((ucount < tv) && (icount < tv))) {
			bpfexpr = "tcp";
			cur_trigger = USETCP;	
		}
		e = trigger_dynbpf_compile(TOPNIP, bpfexpr);
		if (e != NULL) {
			fprintf(fd, "elog: topn_ip trigger parse failure on %s\n", bpfexpr);
			return;
		}
        }

	switch(cur_trigger) {
	case USEICMP:
		tvalue = l->icmpcount/30;
		break;
	case USETCP:
		tvalue = l->tcpcount/30;
		break;
	case USEUDP:
		tvalue = l->udpcount/30;
		break;
	default:
		tvalue = 0;
	}

	trigger_condition_check(TOPNIP, fd, tvalue, "topn_ip"); 
}

/* print flow in semi-human readable format
*/
static void
printFlow(FILE *fd, struct node *head)  
{
	char b1[1024];
	char buf[1024];
	char *ps;

	ps = prototostring(head->ipkey.ipproto);

	if (ps == NULL) {
		sprintf(buf, "%d", head->ipkey.ipproto);
	}
	else
		strcpy(buf, ps);

	sprintf(b1, "%s", intoa(head->ipkey.ipdst));
	fprintf(fd, 
		"%s.%d->%s.%d[%s]: bytes=%u:pkts=%u:size=%u/%u/%u/%u:l7=%u:f=%x:l=%x:r=%x:\n",
		intoa(head->ipkey.ipsrc),
		head->ipkey.srcport,
		b1,
		head->ipkey.dstport,
		buf,  /* protocol */
		head->counter,  /* byte count */
		head->pktcount,
		head->sizes[0], head->sizes[1], head->sizes[2], head->sizes[3],
		head->payload,
		head->flags,
		head->apptags.lower,
		head->apptags.upper 
		);
}

/* print flow in easy to parse for perl format, to blazes with humanity
*/

static void
printPerlFormat(FILE *fd) 
{
	 fprintf(fd, "# ipsrc.l4src->ipdst.l4dst[proto]:bytes:pkts:size0:size1:size2:size3:payloads:flags:applow:apphi:c1:c2:c3\n"); 
}

static void
printPerl(FILE *fd, struct node *head)  
{
	char b1[1024];
	char buf[1024];
	char *ps;

	sprintf(b1, "%s", intoa(head->ipkey.ipdst));
	fprintf(fd, 
		"%s.%d->%s.%d:[%d]:%u:%u:%u:%u:%u:%u:%u:%x:%x:%x:%u:%u:%u:\n",
		intoa(head->ipkey.ipsrc),
		head->ipkey.srcport,
		b1,
		head->ipkey.dstport,
		head->ipkey.ipproto,  /* protocol */
		head->counter,  /* byte count */
		head->pktcount, /* pkt count */
		/* sizes */
		head->sizes[0], head->sizes[1], head->sizes[2], head->sizes[3],
		/* payload counter */
		head->payload,
		/* app flags old */
		head->flags,
		/* pattern matching tags */
		head->apptags.lower,
		head->apptags.upper, 
		head->c1,
		head->c2, 
		head->c3 
		);
}
 
/* print flow in compact binary format - make life interesting for backend
 * this is for benefit of botmesh work and only uses fields of interest
*/
static void
printHex(FILE *fd, struct node *head)  
{
	char b1[1024];
	char buf[1024];
	char *ps;
	struct binformat {
		unsigned int ipsrc;
		unsigned int ipdst;
		unsigned short srcport;
		unsigned short dstport;
		unsigned short ipproto;
		unsigned int bytecount;
		unsigned int pktcount;
		unsigned int size1;
	} bin;

	bin.ipsrc = head->ipkey.ipsrc;
	bin.ipdst = head->ipkey.ipdst;
	bin.srcport = head->ipkey.srcport;
	bin.dstport = head->ipkey.dstport;
	bin.ipproto = head->ipkey.ipproto;
	bin.bytecount = head->counter;
	bin.pktcount = head->pktcount;
	bin.size1 = head->sizes[1];

	fwrite(&bin, sizeof (struct binformat), 1, fd);
}

#ifdef TESTMAIN

static int f = 0;

unsigned int
getRandomWord()
{
	static unsigned int rw;

#define REALLYRANDOM
#ifdef REALLYRANDOM
	if (f == 0) {
		f = open("/dev/urandom", 0);
		if (f < 0) {
			perror("open");
			exit(1);
		}
	}
	read(f, &rw, sizeof(unsigned int));
#else
	rw = rand();
#endif
	return(rw);
}

functionTest()
{
	struct list l;
	COUNTER c;
	struct ipkey ik;
	int rc;
	struct node *n = NULL;

	/* load sample ipkey and counter
	*/
	ik.ipsrc = inet_addr("1.1.1.1");
	ik.ipdst = inet_addr("1.1.1.2");
	ik.srcport = ik.dstport = 0;
	ik.ipproto = IPPROTO_TCP;
	c = 1;	

	initList(&l);

	searchInsertNode(&l, &ik, c);
	
	searchInsertNode(&l, &ik, c);

	searchInsertNode(&l, &ik, c);

	/* add 4 more nodes and dump the list
	*/
	ik.ipsrc = inet_addr("1.1.1.2");
	ik.ipdst = inet_addr("1.1.1.3");
	c = 2;
	searchInsertNode(&l, &ik, c);

	ik.ipsrc = inet_addr("1.1.1.3");
	ik.ipdst = inet_addr("1.1.1.4");
	c = 3;
	ik.ipproto = IPPROTO_UDP;
	searchInsertNode(&l, &ik, c);

	ik.ipsrc = inet_addr("1.1.1.4");
	ik.ipdst = inet_addr("1.1.1.5");
	ik.ipproto = 1;
	c = 4;

	searchInsertNode(&l, &ik, c);

	ik.ipsrc = inet_addr("1.1.1.6");
	ik.ipdst = inet_addr("1.1.1.7");
	c = 5;
	searchInsertNode(&l, &ik, c);

	/* now search for a given node, and increment its
	 * value
	 */
	c = 5;
	searchInsertNode(&l, &ik, c);

	/* now search for a previous node, note last
	 * as a simple cache should not be used here
	 */
	ik.ipsrc = inet_addr("1.1.1.4");
	ik.ipdst = inet_addr("1.1.1.5");
	c = 1;
	searchInsertNode(&l, &ik, c);

	dumpHash(&l);
	printList(stdout, &l, 1, 1, 1, 1, 5);
	destroyList(&l);
}

randomTest()
{
	struct list l;
	struct ipkey ik;
	unsigned int key1, key2;
	COUNTER counter1;
	unsigned int k,c;
	struct node *n;
	int i;
	int rc;

	initList(&l);
	
	for ( i = 0; i < 10000; i++) {
		key1 = getRandomWord();
		key2 = getRandomWord();
		counter1 = getRandomWord();
		ik.ipsrc = key1;
		ik.ipdst = key2;

		printf("[i=%d], key1=%lx, key2=%lx, counter=%lx\n", i, key1, key2, counter1);

		searchInsertNode(&l, &ik, counter1);
		searchInsertNode(&l, &ik, counter1);
	}
	
	printf("node count %d\n", l.count);
	dumpHash(&l);
	printList(stdout, &l, 1, 1, 1, 1, 5);

	destroyList(&l);
}

struct ourstats os;

main()
{
	/*
	functionTest();
	*/
	randomTest();
}
#endif

