/*
 * ircscan.h
 */
#ifndef IRCSCAN_H
#define IRCSCAN_H

#ifndef IPCOUNTER
typedef unsigned int  COUNTER;
#endif

#define ISJOIN  1
#define ISPING  2
#define ISPONG  3
#define ISPRIVMSG 4

#define ISSRC 1
#define ISDST 2

/*
 * basic list data organization
 *
 * channel hashlist
 *	channel   channel  channel
 *	channel
 *
 * irc node hashlist
 *	ipnode	ircnode	ircnode  NULL
 *		ircnode
 *		ircnode
 *
 * each channel node in turn can point to a linked linear list of
 *	ircnodes.  this list is currently searched in a linear fashion.
 *
 * lookup ircnode: via irc node hashlist
 * lookup channode: via channel node hashlist
 *
 * join/privmsg:
 *	ip src/ip dst
 *		lookup ircnode
 *		if exists
 *			increment counters
 *		else
 *			create/put in ircnode list/increment counters
 *	channel name
 *		lookup the channel
 *		if exists
 *			increment counters
 *		else
 *			create/put in channel list/increment counters
 *		for ip src/ip dst
 *			linear search of channel node to find ip node
 *			if not found
 *				insert at front
 *			increment counters
 *			
 * ping/pong:
 *	ip src/ip dst
 *		lookup ircnode for ip address (src OR dst, do both)
 *		if exists
 *			increment counters
 *		else
 *			create/put in ircnode list/increment counters
 */

/*
 * an ircnode - can be chained together in a channel grouping only.
 *
 * 
 */

struct ircnode {
	unsigned int ipsrc;	    /* index */
	COUNTER serverhits;	    /* host has irc server */
	COUNTER joinsrc;	    /* ip src for join */
	COUNTER joindst;	    /* ip dst for join */
	COUNTER pingsrc;	    /* ping ip src */
	COUNTER pingdst;	    /* ping ip dst */
	COUNTER pongsrc; 	    /* pong ip src */
	COUNTER pongdst;	    /* pong ip dst */
	COUNTER privsrc;	    /* privmsg ip src */
	COUNTER privdst;	    /* privmsg ip dst */
	COUNTER nodehits;	    /* synthesized counters */
	COUNTER jointotal;
	COUNTER pingtotal;
	COUNTER pongtotal;
	COUNTER privtotal;
	COUNTER channels;	    /* unique channel count */
	unsigned short dport;       /* destination port - sampled */
	unsigned short sport;       /* src port - sampled */
	struct ircnode *next;       /* hashlist for ips */
};

/* glue pointers for ircnodes
*/
struct irclink {
	struct ircnode *ircnode;
	struct irclink *nextlink;
};

/*
 * basic data structure for channel name.
 * we point to a list of ircnodes via a irclink simple list.
 *
 * not keeping the key for "privacy" protection
 * of course any ASCII password is not terribly private.
 */
struct channel {
	char *name;	/* MALLOCED. channel name unique key */
	COUNTER noips;  /* sort index: total number of ips in channel */
	COUNTER hits;   /* how many times uttered total */
	COUNTER joins;	/* how many of these kinds of messages */
	COUNTER privmsgs;	/* how many of these kinds of messages */
	struct channel *nextchan;   /* lookup: hashlist of channels */
	struct irclink *nextlink;   /* list of ircnodes */
};


/* can be longer but we will heuristically assume 
 * 32 chars as max length
 *
 * note: 1st char either # or & or apparently :
 * end denoted usually by blanks,
 * formal end:
 * [SPACE, COMMA, NUL, Cntrl-G]
 */

#define CHARSPACE	' '
#define CHARCOMMA       ','
#define CHARNUL         0x0
#define CHARBELL        0x7

#define MAXCHANNAME 32

/* a list itself.  it has pointers to the linked-list
 * and a count of elements AND a hash access mechanism
 * for fast searching, based on node keys.
 */
struct irclist {
	unsigned int ipcount;	 /* entries in ip hashlist */
	unsigned int chancount; /* entries in chan hashlist */
	unsigned int totaljoins;  /* total joins */
	unsigned int totalpings;  /* total pings */
	unsigned int totalpongs;  /* total pongs */
	unsigned int totalprivmsgs;  /* total privmsgs */

#define IRCHASH	104729
	struct ircnode *iphash[IRCHASH];
	struct channel *chanhash[IRCHASH];
};

#endif 

void printIrc(FILE *, struct irclist *, struct synlist *, int , char *);
