
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <syslog.h>
#include "patmatch.h"

static int 
__inline
patmatch_matchWrapper(struct patdescr *head, unsigned char *payload, int payloadlen)
{
	int oveccount=30;
	int ovector[oveccount]; //we do not really care about substrings
	int result;
	int rd=0;
	int wr=0;
	char *newpayload;

	if (head->removeNull==YES) {
		newpayload=(char *) malloc(payloadlen);
		if (newpayload == NULL) {
			syslog(LOG_ERR, 
			"patmatch_matchWrapper: malloc failure during NULL payload creation.");
			exit(1);
		}
		for(rd=0;rd<payloadlen;rd++) {
			if (payload[rd] != '\0') {
				newpayload[wr]=payload[rd];
				wr++;
			}
		}
		result=pcre_exec(head->compiled,head->extrainfo,newpayload, wr, 0, 0, ovector, oveccount);
		free(newpayload);
	} else {
		result=pcre_exec(head->compiled,head->extrainfo, (const char *)payload, payloadlen, 0, 0, ovector, oveccount);
	}
	return(result);
}

/*add a pattern, its tag and protocol number to the data
 * structure for patterns. 
 *
 * The following checks and changes are performed on the pattern:
 * 1) white space at the front and after quote at end are removed
 * 2) if there is a quote at the beginning and end then both are removed.
 * 3) if there is only a quote at the beginning (and not at the end) then 
 *	  pattern is discarded.
 *
 * TODO: if the pattern starts with a quote and ends with an escaped quote
 * then the quotes will be removed but the escape character not. hopefully
 * pcre will pick up problems with these kinds of patterns.
 * TODO: handle studied patterns
 */
int patmatch_addPattern(char *proto, char tag, char *pkts, char *bytes, u_short removenull, char *pattern,struct patdescr **head)
{
	u_short			protocol;
	int				result;
	struct patdescr *newpat;
	struct patdescr *pat=*head;
	const char	   *error;
	int				erroffset;
	int maxpkts;
	int maxbytes;
	char const *ep;

	/*first check if tag does not already exist*/
	while (pat !=NULL) {
		if (pat->tag==tag) {
			fprintf(stderr,"patmatch_addPattern: duplicate tag, pattern not added\n");
			return(1);
		}
		pat=pat->next;
	}
	pat=*head;

	if (strcmp(pkts,"DEFAULT") == 0) {
		maxpkts=MAX_PACKETS;
	} else {
		maxpkts=atoi(pkts);
	}

	if (strcmp(bytes,"DEFAULT") == 0) {
		maxbytes=MAX_BYTES;
	} else {
		maxbytes=atoi(bytes);
	}

	/*convert transport protocol name to protocol number*/
	protocol=patmatch_getProto(proto);
	if (protocol == MAX_U_SHORT) {
		fprintf(stderr,"patmatch_addPattern: unknown protocol in pattern with tag %c, pattern not added. \n", tag);
		return(1);
	}

	if (patmatch_patternChecked(&pattern)==FALSE) {
		fprintf(stderr,"patmatch_addPattern: pattern check on tag %c failed, pattern not added.\n",tag);
		return(1);
	}

	if (patmatch_tagChecked(tag)==FALSE) {
	   fprintf(stderr,"patmatch_addPattern: tag %c not allowed, pattern not added.\n",tag);
	   return(1);
	}

	/*checks completed successfully, now add the pattern.*/
	newpat=(struct patdescr *) malloc(sizeof(struct patdescr));
	if (newpat == NULL) {
		fprintf(stderr,"patmatch_addPattern: malloc failure during new pattern add.\n");
		return(1);
	}
	bzero(newpat, sizeof(struct patdescr));
	newpat->tag=tag;
	newpat->protocol=protocol;
	newpat->removeNull=removenull;
	newpat->maxpkts=maxpkts;
	newpat->maxbytes=maxbytes;
	if (strlen(pattern)<MAX_PAT_LEN) {
		strcpy(newpat->pattern,pattern);
	} else {
		fprintf(stderr,"patmatch_addPattern: pattern too big. \n");
		return(1);
	}
	newpat->apptags=patmatch_tagToBits(tag);
	newpat->compiled=pcre_compile(pattern, PCRE_CASELESS, &error, &erroffset, NULL);
	if (newpat->compiled==NULL) {
		fprintf(stderr,"patmatch_addPattern: compile pattern error at %d: %s\n", erroffset, error);
		return(1);
	}
	ep = NULL;
	newpat->extrainfo = pcre_study(newpat->compiled, 0, &ep);
	if (ep) {
		fprintf(stderr,"patmatch_addPattern: pcre_study error for pattern %s\n", pattern);
		return(1);
	}
	newpat->next=NULL;
	if (pat == NULL) {
		*head=newpat;
	} else {
		newpat->next=*head;
		*head=newpat;
	}
	return(0);
}


void patmatch_printPatterns(struct patdescr *head)
{
	while (head !=NULL) {
		printf("Tag: %c\n lower: %d\n upper: %d\n protocol:%s\n maxpkts:%u\n maxbytes:%u\n removeNull:%u\n pattern:%s\n\n",head->tag,head->apptags.lower,head->apptags.upper,patmatch_getProtoName(head->protocol), head->maxpkts, head->maxbytes, head->removeNull, head->pattern);
		head=head->next;
	}
}

void patmatch_printTags(struct tags apptags)
{
	printf("tag values: ");
	if (apptags.lower & aVALUE) { printf("a "); }
	if (apptags.lower & bVALUE) { printf("b "); }
	if (apptags.lower & cVALUE) { printf("c "); }
	if (apptags.lower & dVALUE) { printf("d "); }
	if (apptags.lower & eVALUE) { printf("e "); }
	if (apptags.lower & fVALUE) { printf("f "); }
	if (apptags.lower & gVALUE) { printf("g "); }
	if (apptags.lower & hVALUE) { printf("h "); }
	if (apptags.lower & iVALUE) { printf("i "); }
	if (apptags.lower & jVALUE) { printf("j "); }
	if (apptags.lower & kVALUE) { printf("k "); }
	if (apptags.lower & lVALUE) { printf("l "); }
	if (apptags.lower & mVALUE) { printf("m "); }
	if (apptags.lower & nVALUE) { printf("n "); }
	if (apptags.lower & oVALUE) { printf("o "); }
	if (apptags.lower & pVALUE) { printf("p "); }
	if (apptags.lower & qVALUE) { printf("q "); }
	if (apptags.lower & rVALUE) { printf("r "); }
	if (apptags.lower & sVALUE) { printf("s "); }
	if (apptags.lower & tVALUE) { printf("t "); }
	if (apptags.lower & uVALUE) { printf("u "); }
	if (apptags.lower & vVALUE) { printf("v "); }
	if (apptags.lower & wVALUE) { printf("w "); }
	if (apptags.lower & xVALUE) { printf("x "); }
	if (apptags.lower & yVALUE) { printf("y "); }
	if (apptags.lower & zVALUE) { printf("z "); }
	if (apptags.upper & AVALUE) { printf("A "); }
	if (apptags.upper & BVALUE) { printf("B "); }
	if (apptags.upper & CVALUE) { printf("C "); }
	if (apptags.upper & DVALUE) { printf("D "); }
	if (apptags.upper & EVALUE) { printf("E "); }
	if (apptags.upper & FVALUE) { printf("F "); }
	if (apptags.upper & GVALUE) { printf("G "); }
	if (apptags.upper & HVALUE) { printf("H "); }
	if (apptags.upper & IVALUE) { printf("I "); }
	if (apptags.upper & JVALUE) { printf("J "); }
	if (apptags.upper & KVALUE) { printf("K "); }
	if (apptags.upper & LVALUE) { printf("L "); }
	if (apptags.upper & MVALUE) { printf("M "); }
	if (apptags.upper & NVALUE) { printf("N "); }
	if (apptags.upper & OVALUE) { printf("O "); }
	if (apptags.upper & PVALUE) { printf("P "); }
	if (apptags.upper & QVALUE) { printf("Q "); }
	if (apptags.upper & RVALUE) { printf("R "); }
	if (apptags.upper & SVALUE) { printf("S "); }
	if (apptags.upper & TVALUE) { printf("T "); }
	if (apptags.upper & UVALUE) { printf("U "); }
	if (apptags.upper & VVALUE) { printf("V "); }
	if (apptags.upper & WVALUE) { printf("W "); }
	if (apptags.upper & XVALUE) { printf("X "); }
	if (apptags.upper & YVALUE) { printf("Y "); }
	if (apptags.upper & ZVALUE) { printf("Z "); }
	printf("\n");
}




int patmatch_isEmpty(struct patdescr *head)
{
	if (head==NULL) {
		return(TRUE);
	} else {
		return(FALSE);
	}
}

/*function will receive tags that should not be matched for - potentially
 * these tags have already been associated with the source of the payload and
 * do not need to be matched again
 * returns a tags structure indicating all patterns that were matched 
 * against the payload - including the given tags
 * */
int patmatch_matchAll(u_short protocol, struct tags *apps, char *payload, struct patdescr *head, int payloadlen)
{
	int oveccount=30;
	int ovector[oveccount]; //we do not really care about substrings
	int result;
	while (head != NULL) {
		/*printf("matchAll: testing pattern with tag %c, protocol %d\n", head->tag, protocol);*/
		/*only match patterns with the same transport protocol as the payload  
		 * under consideration*/
		if (head->protocol == protocol) {
			/*printf("matchAll: protocol same as payload.\n");*/
			/*check if tag is associated with the source of this payload*/
			if (((apps->lower & head->apptags.lower) != 1)&&((apps->upper & head->apptags.upper) != 1)) {

				/*printf("matchAll: pattern never matched before.\n");*/
				result=pcre_exec(head->compiled,head->extrainfo,payload, payloadlen, 0, 0, ovector, oveccount);

				/*printf("matchAll: match result=%d\n",result);*/
				if (result >= 0) {
					/*pattern matched, now update apptags*/
					apps->lower |=head->apptags.lower;
					apps->upper |=head->apptags.upper;
				} else {
					if (result != PCRE_ERROR_NOMATCH) {
						fprintf(stderr,"patmatch_matchAll: match error: %d\n", result);
						return(1);
					} //else {
						/*printf("matchAll: pattern did not match payload.\n");*/
					/*}*/
				}
			}
		}
		head=head->next;
	}
	return(0);
}

int patmatch_matchWithHistory(u_short protocol, struct tags *apps, unsigned char *payload, struct patdescr *head, int payloadlen, struct tags *attempted, struct tags *appres, unsigned int pkts, unsigned int bytes)
{
	int result;

	while (head != NULL) {
		/*printf("matchWithHistory: testing pattern with tag %c, protocol %d\n", head->tag, protocol);*/
		/*printf("Tag: %c\n lower: %d\n upper: %d\n protocol:%s\n maxpkts:%u\n maxbytes:%u\n pattern:%s\n\n",head->tag,head->apptags.lower,head->apptags.upper,patmatch_getProtoName(head->protocol), head->maxpkts, head->maxbytes, head->pattern);*/
		/*only match patterns with the same transport protocol as the payload  
		 * under consideration*/
		if (head->protocol == protocol) {
			if (((head->maxpkts==0)&&(head->maxbytes==0))||((pkts<head->maxpkts)&&(bytes<head->maxbytes))) {
			/*printf("matchWithHistory: protocol same as payload.\n");*/
			/*check if tag is already associated with the source of this payload*/
			if (((apps->lower & head->apptags.lower) == head->apptags.lower)&&((apps->upper & head->apptags.upper) == head->apptags.upper)) {
				/*printf("matchWithHistory: pattern matched in main structure. no matching performed.\n");*/
			} else {
				/*printf("matchWithHistory: pattern never matched before.\n");*/
				/*check if previous caller of this routine searched for this pattern*/
				if (((attempted->lower & head->apptags.lower) == head->apptags.lower)&&((attempted->upper & head->apptags.upper) ==head->apptags.upper)) {
					/*pattern was searched for before, get result from previous 
					 * matches*/
					/*printf("matchWithHistory: pattern matched on previous attempt, copying the result. \n");*/
					apps->lower |=appres->lower;
					apps->upper |=appres->upper;
				} else {
					/*printf("matchWithHistory: pattern never attempted on this payload. attempting it now\n");*/
					result=patmatch_matchWrapper(head,payload, payloadlen);
					/*result=pcre_exec(head->compiled,head->extrainfo,payload, payloadlen, 0, 0, ovector, oveccount);*/
					attempted->lower |=head->apptags.lower;
					attempted->upper |=head->apptags.upper;
					/*if (result > 0) {printf("matchWithHistory: match result=%d\n",result); }*/
					if (result >= 0) {
						/*pattern matched, now update apptags*/
						appres->lower |=head->apptags.lower;
						appres->upper |=head->apptags.upper;
						apps->lower |=head->apptags.lower;
						apps->upper |=head->apptags.upper;
					} else {
						if (result != PCRE_ERROR_NOMATCH) {
							fprintf(stderr,"patmatch_matchWithHistory: match error: %d\n", result);
							return(1);
						} //else {
							/*printf("matchWithHistory: pattern did not match payload.\n");*/
						/*}*/
					}
				}
			}
		}
		}
		head=head->next;
	}
	return(0);
}
/*private functions */

/*function is provided with the string describing a protocol on top of IP
 * the integer protocol value, defined in patmatch.h is returned*/
u_short patmatch_getProto(char *proto) 
{
	if (strcmp(proto,"TCP") == 0) {
	   return(IPPROTO_TCP);
	} else if (strcmp(proto,"UDP") == 0) {
		return(IPPROTO_UDP);
	} else if (strcmp(proto, "ICMP") == 0) {
		return(IPPROTO_ICMP);
	} else {
		return(MAX_U_SHORT);
	}
}

/*given the integer value of a transport protocol this function returns the 
 * string value*/
char *patmatch_getProtoName(u_short proto)
{
	char *protoname;
	if (proto == IPPROTO_TCP) {
		protoname="TCP";
	} else if (proto == IPPROTO_UDP) {
		protoname="UDP";
	} else if (proto == IPPROTO_ICMP) {
		protoname="ICMP";
	} else {
		protoname="UNKNOWN";
	}
	return(protoname);
}


/*given a character tag, convert it into the tags structure*/
struct tags patmatch_tagToBits(char tag)
{
	struct tags newtags;

	bzero(&newtags,sizeof(newtags));
	switch (tag) {
		case 'a':
			newtags.lower=aVALUE;
			break;
		case 'b':
			newtags.lower=bVALUE;
			break;
		case 'c':
			newtags.lower=cVALUE;
			break;
		case 'd':
			newtags.lower=dVALUE;
			break;
		case 'e':
			newtags.lower=eVALUE;
			break;
		case 'f':
			newtags.lower=fVALUE;
			break;
		case 'g':
			newtags.lower=gVALUE;
			break;
		case 'h':
			newtags.lower=hVALUE;
			break;
		case 'i':
			newtags.lower=iVALUE;
			break;
		case 'j':
			newtags.lower=jVALUE;
			break;
		case 'k':
			newtags.lower=kVALUE;
			break;
		case 'l':
			newtags.lower=lVALUE;
			break;
		case 'm':
			newtags.lower=mVALUE;
			break;
		case 'n':
			newtags.lower=nVALUE;
			break;
		case 'o':
			newtags.lower=oVALUE;
			break;
		case 'p':
			newtags.lower=pVALUE;
			break;
		case 'q':
			newtags.lower=qVALUE;
			break;
		case 'r':
			newtags.lower=rVALUE;
			break;
		case 's':
			newtags.lower=sVALUE;
			break;
		case 't':
			newtags.lower=tVALUE;
			break;
		case 'u':
			newtags.lower=uVALUE;
			break;
		case 'v':
			newtags.lower=vVALUE;
			break;
		case 'w':
			newtags.lower=wVALUE;
			break;
		case 'x':
			newtags.lower=xVALUE;
			break;
		case 'y':
			newtags.lower=yVALUE;
			break;
		case 'z':
			newtags.lower=zVALUE;
			break;
		case 'A':
			newtags.upper=AVALUE;
			break;
		case 'B':
			newtags.upper=BVALUE;
			break;
		case 'C':
			newtags.upper=CVALUE;
			break;
		case 'D':
			newtags.upper=DVALUE;
			break;
		case 'E':
			newtags.upper=EVALUE;
			break;
		case 'F':
			newtags.upper=FVALUE;
			break;
		case 'G':
			newtags.upper=GVALUE;
			break;
		case 'H':
			newtags.upper=HVALUE;
			break;
		case 'I':
			newtags.upper=IVALUE;
			break;
		case 'J':
			newtags.upper=JVALUE;
			break;
		case 'K':
			newtags.upper=KVALUE;
			break;
		case 'L':
			newtags.upper=LVALUE;
			break;
		case 'M':
			newtags.upper=MVALUE;
			break;
		case 'N':
			newtags.upper=NVALUE;
			break;
		case 'O':
			newtags.upper=OVALUE;
			break;
		case 'P':
			newtags.upper=PVALUE;
			break;
		case 'Q':
			newtags.upper=QVALUE;
			break;
		case 'R':
			newtags.upper=RVALUE;
			break;
		case 'S':
			newtags.upper=SVALUE;
			break;
		case 'T':
			newtags.upper=TVALUE;
			break;
		case 'U':
			newtags.upper=UVALUE;
			break;
		case 'V':
			newtags.upper=VVALUE;
			break;
		case 'W':
			newtags.upper=WVALUE;
			break;
		case 'X':
			newtags.upper=XVALUE;
			break;
		case 'Y':
			newtags.upper=YVALUE;
			break;
		case 'Z':
			newtags.upper=ZVALUE;
			break;
		default:
			/*will return tags structure with all zeros*/
			fprintf(stderr,"patmatch_addpattern: unknown tag value.\n");
			break;
		}
		return(newtags);
}


/*do basic sanity check on regular expression provided and remove 
 * matching quotes around pattern*/
int patmatch_patternChecked(char **pattern)
{
	char *pat=*pattern;
	int len;
	/*remove white space from the front of the pattern*/
	while (*pat == ' ' || *pat == '\t') {
		pat++;
	}
	len=strlen(pat);

	/*remove quotes from pattern if present if a pattern starts with a quote
	* then the last character also has to be a quote. if there is a
	* quote at the end of the pattern, but a space afterwards, then the
	* spaces are removed first */
	if (pat[0] == '"') {
		/*first see if we can remove whitespace at end of pattern to reach
		* a matching quote*/
		while ((len > 0) && (pat[len-1]== ' '|| pat[len-1]== '\t')) {
		   pat[len-1]='\0';
		   len--;
		}
		if (pat[len-1]=='"' ) {
			pat[len-1]='\0';
			pat++;
		} else {
			fprintf(stderr,"patmatch_patternChecked: no matching quotes\n");
			return(FALSE);
		}
	}

	/*do not accept empty patterns. note, pattern could be made empty in
	* search of matching quote*/
	if (len == 0) {
		fprintf(stderr,"patmatch_patternChecked: empty pattern.\n");
		return(FALSE);
	}

	*pattern=pat;
	/*printf("patmatch_patternChecked: added pattern: %s\n",pat);*/
	return(TRUE);
}

/*a few tags are currently reserved - they will be used in hardcoded searches
 * in hashsyn.c
 * I - irc
 * P - honeynet ip dest addr*/
int patmatch_tagChecked(char tag) 
{
	if (tag=='I'||tag=='P') {
		return(FALSE);
	}
	return(TRUE);
}



