/*
 * trigger.c - trigger capability 
 *
 * This file includes functions of trigger configuration and condition
 * checking , and trigger_work. Triggers are used to capture and store packets if ourmon 
 * detects anomalies. 
 *
 * The working procedure for a trigger is:
 *
 * configure the trigger if it is specified in ourmon.conf
 * check the trigger condition when it is writing output.
 * if condition satisfied, trigger starts working and captures the specified 
 * number of packets during non-write-output packet capture.
 * The trigger is turned off when the desired number of packets are captured.
 */

/*
 * Copyright (c) 2004 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse
 *    or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <signal.h>
#include <sys/stat.h>
#include <string.h>
#include <time.h>
#include <pcap.h>

#include "trigger.h"
#include "ourmon.h"
#include "filter.h"
#include "sync.h"

#ifndef TRUE
#define TRUE    1
#endif
#ifndef FALSE
#define FALSE   0
#endif

extern pcap_t *pd;

int trigger_count = 0;		/* trigger configured */
/* global trigger state.  If > 1 then some trigger
 * is set
 */
unsigned int trigger_worm_state=0;
struct trigger_cap trigger_cap[MAXTRIGGERS]; /* the list of triggers */

/* 
 * basic information about trigger types
 * type, state value, name, default bpf expression
 */ 
const struct trigger_info trigger_info[] = {
	{TWORM, 0x01, "tworm", "tcp[tcpflags] & tcp-syn != 0"},
	{TOPNIP, 0x02, "topn_ip", "ip"}, 
	{TOPNSCANS, 0x03, "topn_scan", ""},
	{TOPNUDPERR, 0x04, "topn_udp_err", ""},
	{DROP, 0x05, "drop_packets", "ip"},
        {TOPNUDPEVIL, 0x06, "topn_udp_evil", ""},
	{0, 0, NULL, NULL}, 
};

int bpf_info_last=0;

/* a table to store the trigger capability index
 * for any graph/label bpf expression used in ourmon
 * The value stored here is simply the trigger capability
 * index.
 */
int bpf_trigger_cap[MAXBPFSPECS][MAXBPFS];


static void trigger_bpf_compile(struct bpf_program *fcode,	/* output */
	int bpfop, int snaplen, unsigned int netmask, 
	char* bpfexpr /* input */);

/*
 * get system timestamp for constructing
 * unique trigger_on time filename
 */
void 
get_ts(char *ts)
{
	struct tm *tm;
	time_t t;
	
	t = time(NULL);
	tm = localtime(&t);
	sprintf(ts,"%02d.%02d.%04d.%02d:%02d:%02d",
			tm->tm_mon+1,
			tm->tm_mday,
			tm->tm_year + 1900,
			tm->tm_hour,
			tm->tm_min,
			tm->tm_sec);
}

/*
 * initialize a trigger capability
 *
 * called at config time, therefore can
 * do a printf on an error.  If this idea is
 * not the case, error i/o needs to change.
 *
 * goal: store the value, count, name in the trigger
 *	structure.
 * input:
 *	type:	trigger type
 *	value:	thresh value
 *	count:  number of packets intended to capture
 *	name:	directory where the packet file is saved.
 * returns:
 *	0 - ok
 *     -1 - error
 */
int 
trigger_config(int type, char *value, char *count, char *name, int multiplier)
{
	int i;
	char *dfile;
	struct stat statbuf;
	extern  int errno;

	if( type >= MAXTRIGGERS) {
		fprintf(stderr,
			"trigger_config called with type %d - too many triggers\n");
		return (-1); 
	}

	/* check the existence of directory */ 
	if( stat(name, &statbuf) < 0 ) {
		fprintf(stderr,"trigger_config: stat of %s failed: error: %s\n",
			name, strerror(errno));
		return (-1); 
	}
	if( S_ISDIR(statbuf.st_mode) == 0 )  {
		fprintf(stderr,"trigger_config: %s is not a directory\n",name);
		return (-1);
	}

	/* note: type is just an array index that is 
	 * determined manually at this point
	 */
	trigger_cap[type].N_value = atoi(value);
	/* caller decides if per second or per period 
	 * if per second we multiply by 30 seconds for period time.
	*/
	if (multiplier) {
		trigger_cap[type].N_value *= multiplier;	
	}
	trigger_cap[type].count = atoi(count);
	if (trigger_cap[type].count < 0) {
               fprintf(stderr,"trigger: count cannot be negative\n");
               return (-1);
	}
        if (trigger_cap[type].N_value < 0) {
               fprintf(stderr,"trigger: value cannot be negative\n");
               return (-1);
        }
        if(strlen(name) > (1022-1)) {
              fprintf(stderr,"trigger: dump file path too long\n");
              return (-1);
        }

        trigger_cap[type].configured = TRUE;
	
	/* form the tag name for the trigger.  this follows
	 * the directory supplied in the config
	*/
	if(type < BPFTRIGGER ) { 
		dfile = trigger_info[type].name;
	}
	else {
		dfile = "bpf";
	}

	/* form the config time base filename, which is used
	 * at runtime to construct a dynamically instanced unique filename
	*/
	if( (strlen(name)+strlen(dfile)) > (2048-1) ){
		fprintf(stderr,"trigger:dump file name too long\n");
		return (-1);
	}
        sprintf(trigger_cap[type].base_filename, "%s/%s", name, dfile);

	/* trigger_count is used in trigger_work 
	 * avoid walk through the whole array. 
	 * value is the maximum of the number of configured trigger 
	 * and the type(index) of trigger which has been configured 
	 * Since bpf trigger cap has its own way 
	 * and does not call trigger_work.
 	 * so trigger_count only refers to #any other trigger cap
	 */    
	if( type <  BPFTRIGGER  )  {
		trigger_count = ( trigger_count >= type ) ? 
				( trigger_count + 1 ):type; 
	}

	return 0;
}

/*
 * trigger_bpf_config
 *
 * Given a trigger capability structure, with trigger_config already
 * called.  Compile a bpf expression for determining
 * whether or not packets should be stored in the trigger file.
 *
 * This routine is called at monconfig time to give a trigger
 * capability a default bpf and needed bpf information.
 *
 * inputs: 
 *	bpfop, snaplen, netmask -- typical bpf params needed
 *	at config time.  
 *
 *	bpfexpr - the bpf expression, which more than likely is
 * 		in the default trigger_info struct above.
 */
int
trigger_bpf_config(int type, int bpfop, int snaplen, 
	int netmask, char *bpfexpr)
{
	if( trigger_cap[type].configured == FALSE ) {
		fprintf(stderr,"trigger_bpf: %s need trigger configured\n",
				trigger_info[type].name);
		return (-1);		 
	}

	if( type  >= BPFTRIGGER ) {
		fprintf(stderr,"trigger_bpf: type %d configured elsewhere\n");
		return (-1);
	}

	/* save these bpf compile params in case they are needed
	* for dynamic instance
	*/
	trigger_cap[type].bpfop = bpfop;
	trigger_cap[type].snaplen = snaplen;
	trigger_cap[type].netmask = netmask;

	/* no bpfexpr in trigger_info means it will be dynamically compiled*/
	if( trigger_info[type].bpfexpr[0] == '\0' ) {
		return 0;
	}

	/* no particular bpf expression needed so just use default
	 * which is already stored in local trigger_info structure.
	*/

	if( bpfexpr == NULL ) {
		bpfexpr = trigger_info[type].bpfexpr;
	}

	trigger_bpf_compile(
			&trigger_cap[type].fcodes, 
			bpfop, 
			snaplen, 
			netmask, 
			bpfexpr);

	return 0;
}

/* find an available trigger capability
 * for the bpf type
 */
int 
bpf_allo_trigger_cap(void)
{
	int i, nRet=-1;

	/* allocate a trigger in the trigger capability
	*/
	for( i = BPFTRIGGER; i < MAXTRIGGERS; i++) {
		if( trigger_cap[i].configured == FALSE) {
			nRet = i;
			break;
		}
	}
	if( i >= MAXTRIGGERS ) {
		fprintf(stderr,"trigger_cap exhausted\n");
		nRet = -1; 
	}

	return nRet;	
}

/* bpf_trigger_completion
 *
 * complete the trigger capability for bpf expression. 
 * including assignment of file's name where the trigger
 * stores packets with the graph name and the program label name, 
 * the fcodes by copying the already compiled 
 * part in bpfs, and re-calculation N_value if count by bytes.
 * record 'type' (position in trigger capabilities) in bpf_trigger_cap 
 * 	
 *	input:
 *		type, label, blabel, bpfpacket
 */  
int 
bpf_trigger_completion(int type, char *label, char *blabel)
{
	int i,j; 
	int nRet = -1;
	struct bpf_spec *b;
	extern struct bpf_spec bpfs[MAXBPFSPECS];

	/* find the bpf expression whose graph name is 'label' and 
	 * program name is 'blabel'
	 */ 
	b = &bpfs[0];
	for( i = 0; i < MAXBPFSPECS; i++, b++ ) {
		/* if graph label matches, try the line labels
		*/
		if(strncmp(label, b->label, strlen(label)) == 0) {
			for( j = 0; j < b->prog_count; j++) {
				if(strncmp(blabel, b->prog_label[j], strlen(blabel)) == 0 ) {
					char bname[1024];

					/* bpf config time result will be: dirname/bpf_graphtag_labeltag
					*/
					memset(bname, 0, sizeof(bname));
					
					sprintf(bname,"_%s_%s", label,blabel);
					/* finish off bpf config time name
					*/
					strcat(trigger_cap[type].base_filename, bname);

					/* must adjust count to reflect back-end point of view.
					 * mon.lite counts things in 30 second summaries.  back-end
					 * graphics displays bytes per sec or pkts per sec.  must
					 * adjust here and therefore remove a write_report time 
					 * needed calculation.  
					 *
					 * if bytes multiply by 30 and divide by 8
					 * if packets multiply by 30
					*/
					if(b->bpfpackets == 0) {
						trigger_cap[type].N_value = (trigger_cap[type].N_value * 30) / 8;
					}
					else {
						trigger_cap[type].N_value = (trigger_cap[type].N_value) * 30;

					}
#ifdef TRIG_DEBUG
					syslog(LOG_ERR, "trigger_cap[bpf].N_value set to %d\n", 
							trigger_cap[type].N_value);
#endif
					/* record index position */
					bpf_trigger_cap[i][j] = type;
					nRet = 0;
					break;
				}		
			} 
			if( j > b->prog_count ) {
				fprintf(stderr, "bpf program not found!\n");
				trigger_cap[type].configured = FALSE;
			}
			break;
		} 

	}
	return nRet;
}

/*
 * actually do the bpf expression compile
 * This function is designed to be called from monconfig.
 */
static
void 
trigger_bpf_compile(struct bpf_program *fcodes,	/* output */
	int bpfop, int snaplen, unsigned int netmask, 
	char* bpfexpr /* input */)
{
	int i,gotone=0;
	extern int debug;
 
	/* compile the bpf
	*/
	(void) 
	local_pcap_compile_nopcap(snaplen, bpfop, fcodes, bpfexpr, 0, netmask, 1);
} 

/* function for dynamic bpf expression compile
 *
 * assumptions:
 *	monconfig made call to set up static bpf info.
 * returns:
 *	string message for elog.  does NOT exit.
 *	NULL - ok.
*/
char *
trigger_dynbpf_compile(int type, 
	char* bpfexpr /* input */)
{
	static char sbuf[2048];
	int rc;

	if( trigger_cap[type].configured == FALSE ) {
		sprintf(sbuf, "trigger_dynbpf_compile: runtime error for type %s: expr %s\n",
				trigger_info[type].name, bpfexpr);
		return (sbuf);
	}
	
	/* compile the bpf
	*/
	rc = 
	local_pcap_compile_nopcap(
		trigger_cap[type].snaplen, 
		trigger_cap[type].bpfop, 
		&trigger_cap[type].fcodes, 
		bpfexpr, 0, 
		trigger_cap[type].netmask, 
		0);

	if (rc < 0) {
		sprintf(sbuf, "trigger_dynbpf_compile: runtime pcap compile error on %s\n",
			bpfexpr);
		return(sbuf);
	}

	return(NULL);
}

/* turn a trigger capability on
 *
 * input: type
 * output: fname if open worked.  
 *	error string 
 *
 * returns:
 *	0 -  no error.
 *	-1 - error.
*/
static int  
trigger_on(int type, char *fname, char *sbuf)
{
	int i;
	unsigned int state;
	char *tri_name;
	char ts[100];
	char filename[1024];		

#ifdef TRIG_DEBUG
	syslog(LOG_ERR, "trigger_on: entering, type %d\n", type);
#endif

	/* if normal, not user bpf trigger
 	 * set the state value in trigger_worm_state
	*/
	if( type < BPFTRIGGER ) {
		/* get the trigger instance name
  	 	*/
		tri_name = trigger_info[type].name; 
#ifdef TRIG_DEBUG
		syslog(LOG_ERR, "trigger_on: %s:trigger_worm_state before:%d",
				tri_name, trigger_worm_state); 
#endif 
		/* get per trigger state 
	 	* for those not bpf type trigger 
		*/
		state = trigger_info[type].state;
		/* set corresponding trigger in global state variable
		*/
		trigger_worm_state |= state;
#ifdef TRIG_DEBUG
		syslog(LOG_ERR, "trigger_on: %s: trigger_worm_state after: %d", 
				tri_name, trigger_worm_state);
#endif
	}

	/* make a dynamic filename with a timestamp in it
	*/
	bzero(ts, 100);
	bzero(filename, 1024);
	
	get_ts(ts);
	if( strlen(ts) == 0 ) {
		sprintf(sbuf,"trigger_on: %d timestamp failed\n",type);
		return (-1);	
	}

	/* filename = typename.<timestamp>.dmp
	*/
	sprintf(filename,"%s.%s.dmp",
		trigger_cap[type].base_filename,ts);

	/* set the per struct_trigger value to on
	*/
	trigger_cap[type].state = 1;

	/* only capture count packets, set the count
	 */
	trigger_cap[type].curcount = trigger_cap[type].count; 

	/* associate the pcap desc with the file
	*/
	if((trigger_cap[type].pd_dump=pcap_dump_open(pd,filename))
			 == NULL) {
		sprintf(sbuf,
			"trigger_on:pcap_dump_open failed: %s\n",	
			pcap_geterr(pd));
		return (-1); 
	}

	/* set flag to indicate that pcap_open has occurred, 
	 * at some point pcapdone should be done once
	 */
	trigger_cap[type].pcapdone = 0;

	/* 
	 * set flag to indicate that pcap descriptor is not availabe
	 * any more. 
	 */
	trigger_cap[type].pcapstop = 0;
			
	/* return the filename for elog use
	*/
	strncpy(fname, filename, strlen(filename));

	/* save the filename just in case, although we do not use it barring debug
	*/
	trigger_cap[type].inst_filename[0] = 0;
	strncpy(trigger_cap[type].inst_filename, fname, strlen(fname));

#ifdef TRIG_DEBUG
	syslog(LOG_ERR, "trigger_on: dynamic dump file: %s\n",fname);
	syslog(LOG_ERR, "trigger_on: count of #packets %d", trigger_cap[type].count);
	syslog(LOG_ERR, "trigger_on: pcapdone = %d\n", trigger_cap[type].pcapdone);
#endif
	return 0;	
}


/*
 * if the trigger is on, turn it off
 */
void
trigger_off(int type)
{
	unsigned int state=0;

#ifdef TRIG_DEBUG
	syslog(LOG_ERR, "trigger_off %s called, type: %d \n", 
			trigger_cap[type].inst_filename, type);
	syslog(LOG_ERR, "trigger_off: trigger_worm_state before:%d",
				trigger_worm_state);
#endif	

	if( type  < BPFTRIGGER ) {
		state = trigger_info[type].state;
	}

	trigger_worm_state &= ~state;

	/* if pcap already closed due to trigger count
	 * running down, then don't close it again
	 */
	if (trigger_cap[type].pcapdone == 0) {
		trigger_cap[type].pcapdone = 1;
		pcap_dump_close(trigger_cap[type].pd_dump);
	}

	/* reset pcapstop TBD */
	trigger_cap[type].pcapstop = 0;

	trigger_cap[type].curcount = 0;
	trigger_cap[type].state = 0;

#ifdef TRIG_DEBUG
	syslog(LOG_ERR , "trigger_off: %d reset,trigger_worm_state: %d\n",
					type, trigger_worm_state);
	syslog(LOG_ERR, "trigger_off: trigger_cap[%d].state: %d",
					 type, trigger_cap[type].state);
#endif
}

/*
 * 
 * check_pcap_desc
 *	purpose:
 *	check if pcap desciptor is still available for this trigger
 *	in case crash caused by writing to invalid descriptor happens
 * input:
 *	pdump - pcap desc for the trigger
 * return:
 *	0  - ok
 *	-1 - error
 */
static int 
check_pcap_desc(pcap_dumper_t *pdump) 
{
	int ret = 0;
	
	if( ferror((FILE *) pdump) != 0)  {
		syslog(LOG_ERR,"check_pcap_desc: pcap desc error (errno: %m).\n");
		ret = -1;
	}
	
	return ret;
}

/* actually store a trigger packet
*/
int 
trigger_work(struct pcap_pkthdr *h, u_char *p, int caplen, 
	 int packet_len)
{
	int i,j;
	struct bpf_program *b;
	pcap_dumper_t *pdump=NULL;
		
	/* make this work more efficiently by not looping through
	 * all capabilities because you know only the first N were used.
	 * i.e., track the number of capabilities actually used at config time.
	 */
	/* walk through and find 'on' trigger */
	for( i = 0; i <= trigger_count; i++) {
		if ( trigger_cap[i].state == 1) {
			/* if the count is exhausted, don't store any packets
			 * if the pcap file is still open, close it.
			*/
			if (trigger_cap[i].curcount <= 0) {
#ifdef THREAD
				spinlock(&locks[LOCK_TRIGGER]);
#endif
				if (trigger_cap[i].pcapdone == 0)  {
					pcap_dump_close(trigger_cap[i].pd_dump);
					trigger_cap[i].pcapdone = 1;

#ifdef TRIG_DEBUG
					syslog(LOG_ERR, "trigger_work: type: %d, count exhausted, pcap close\n", i);
#endif

				}
#ifdef THREAD
				unlock(&locks[LOCK_TRIGGER]);
#endif
				continue;
			}
			/* 
 			 * if pcap desc has been unavaibable,
			 * no need to save the packet in this file any more
			 */
			if(trigger_cap[i].pcapstop == 1) {
#ifdef THREAD
				spinlock(&locks[LOCK_TRIGGER]);
#endif
				pcap_dump_close(trigger_cap[i].pd_dump);
#ifdef THREAD
				unlock(&locks[LOCK_TRIGGER]);
#endif
				continue;
			}

			/* bpf filters */
			b = &(trigger_cap[i].fcodes);
#ifdef THREAD
			spinlock(&locks[LOCK_TRIGGER]);
#endif
			if( bpf_filter(b->bf_insns, p, 
				packet_len, caplen) == 0 ) {
				/*failure to match*/
				;
			} else {
				/* note that libpcap does not return an error.
				*/
				pdump = trigger_cap[i].pd_dump;
				if(pdump == NULL) {
					syslog(LOG_ERR,"trigger_work: pdump NULL: write file failed\n");
#ifdef THREAD
					unlock(&locks[LOCK_TRIGGER]);
#endif
					continue;
				}
				pcap_dump((u_char*)pdump, h, p);
				trigger_cap[i].curcount--;
			} /*end if(bpf_filter)*/			
#ifdef THREAD
			unlock(&locks[LOCK_TRIGGER]);
#endif
		}  /* if trigger cap. state is on */ 
	} /*end for*/
} 

/*
 * user bpf trigger function. write out a packet
 * in pcap format.
 *
 * There are no instances of pcap_dump called
 * outside this module.
 *
 * LOCKING: this is called within a bpf lock.
 */
inline void
bpf_trigger_work(struct pcap_pkthdr *h, u_char *p, int index)
{
	pcap_dumper_t *pdump;

	/* pcap desc for this trigger */
	pdump = trigger_cap[index].pd_dump;

	if((trigger_cap[index].state == TRUE)) {
		/* TBD. what happens here
		*/
		if (trigger_cap[index].pcapstop == 1) {
			/* pcap desc invalid */
#ifdef TRIG_DEBUG	
			syslog(LOG_ERR,"pcap desc invalid\n");
#endif	
			return;
		}
		if (trigger_cap[index].curcount > 0 ) {
			if(pdump == NULL) {
				syslog(LOG_ERR,"bpf_trigger_work: file descriptor is null.\n");
				return;			
			}
			pcap_dump((u_char*)pdump, h, p);
			if( check_pcap_desc(pdump) < 0) 
				trigger_cap[index].pcapstop = 1; 
			trigger_cap[index].curcount--;
		}
		else {
			if (trigger_cap[index].pcapdone == 0) {
#ifdef TRIG_DEBUG
		syslog(LOG_ERR, "bpf_trigger_work: count exhausted\n");
#endif
				pcap_dump_close(trigger_cap[index].pd_dump);
				trigger_cap[index].pcapdone = 1;
			}
		}
	} /* if state on */
}

/*
 * user bpf function: check trigger state
 *
 * this is called at write_report time.
 */
void
bpf_check_trigger_state(FILE *fd, int candidate, int index)
{
	char fname[1024];
	int ret=0;
	char e[2048];

	bzero(e, 2048);
	if((trigger_cap[index].state == FALSE) &&
		 candidate >= trigger_cap[index].N_value) {

		memset(fname, 0, 1024);
		ret = trigger_on(index, fname, e);
		if(ret != 0) {
#ifdef TRIG_DEBUG
			syslog(LOG_ERR,"elog: bpf trigger_on failed %s\n", e);
#endif
			fprintf(fd,"elog: bpf trigger_on failed %s\n", e);
		}
		else {
#ifdef TRIG_DEBUG
			syslog(LOG_ERR,"elog: bpf trigger on: current value %d: dumpfile: %s\n", candidate, fname);
#endif
			fprintf(fd,"elog: bpf trigger on: real current value/real trigger %d/%d: dumpfile: %s\n", 
					candidate, trigger_cap[index].N_value, fname);
		}
	}
	else if( (trigger_cap[index].state == TRUE) &&
	     (candidate < trigger_cap[index].N_value) ) {
		trigger_off(index);
#ifdef TRIG_DEBUG
		syslog(LOG_ERR,"elog: bpf trigger off, %d/%d\n", candidate, trigger_cap[index].N_value);
#endif
		fprintf(fd,"elog: bpf trigger off, %d/%d\n", candidate, trigger_cap[index].N_value);
	}
}


/*
 * see if trigger should be armed
 *
 * returns:
 *	threshold value if trigger should be armed.
 *	0 - trigger should not be set.
 */

int
pre_trigger_check(int type, int current_value)
{
	/* trigger is configured, state is false, and current value is greater
	 * than trigger threshold.  if so, returned threshold value
	 */
        if (trigger_cap[type].configured && 
		(trigger_cap[type].state == FALSE) && 
                (current_value >= trigger_cap[type].N_value)) { 
#ifdef TRI_DEBUG
		syslog(LOG_ERR, "pre_trigger_check: %d configured and state FALSE\n",
					type);
#endif
		return(trigger_cap[type].N_value);
	}
	return(0);
}

/*
 * perform common logic to turn trigger on or off
 *
 * assumption: this function is called by a trigger user
 *	at write_report time.
 *
 * note: we assume trigger-time bpf expression has already been set either
 *	at compile time or before this function is called.
 */
void
trigger_condition_check(int type, FILE *fd, int current_value, char *tag)
{
	char fname[1024];
        char e[2048]={0,}; 
	int ret = 0;	

#ifdef TRIG_DEBUG
/*
	syslog(LOG_ERR, "trigger_condition_check:%s entering,type: %d\n",
				trigger_cap[type].inst_filename,type);
	syslog(LOG_ERR, "trigger_condition_check:type: %d, current count:%d\n",
				type, current_value);
*/
#endif

        if (trigger_cap[type].configured) {
                if(current_value >= trigger_cap[type].N_value) {

			/* if the trigger is not on, turn it on
			*/ 
			if (trigger_cap[type].state == FALSE) {
                        	memset(fname, 0, sizeof(fname));
                        	ret = trigger_on(type, fname, e);

#ifdef TRIG_DEBUG
				syslog(LOG_ERR, 
					"trigger_condition_check: state FALSE->TRUE. packet file: %s, current count: %d, N_value %d\n", 
					fname, current_value, trigger_cap[type].N_value);
#endif
				if (ret != 0) {
					fprintf(fd,"elog: %s trigger failed: %s\n", tag, e);
				}
				else {
#ifdef TRIG_DEBUG
					syslog(LOG_ERR, "elog ON message issued: (%s) %s %d %d\n", 
						tag, fname, current_value, 
						trigger_cap[type].N_value);
#endif
					fprintf(fd, 
						"elog: %s trigger on, current count: %d, threshold %d, dumpfile: %s\n", 
						tag, 
						current_value, 
						trigger_cap[type].N_value,
						fname);
				}
			}
                }
                else {
			if (trigger_cap[type].state == TRUE) {
#ifdef TRIG_DEBUG
				syslog(LOG_ERR, 
				"trigger_condition_check: state TRUE->FALSE. current count: %d, N_value %d\n", 
					current_value, trigger_cap[type].N_value);
#endif
				trigger_off(type);
#ifdef TRIG_DEBUG
				syslog(LOG_ERR, "elog OFF message issued\n");
#endif
				fprintf(fd, 
				"elog: %s trigger OFF, current count is %d, threshold: %d\n", 
					tag, current_value, trigger_cap[type].N_value);
			}
                }
        }
}          

/* at write report time, make sure all the files
 * are flushed out if any are open
 */
void
trigger_flush()
{
	int i;
		
	/* make this work more efficiently by not looping through
	 * all capabilities because you know only the first N were used.
	 * i.e., track the number of capabilities actually used at config time.
	 */
	/* walk through and find 'on' trigger */
	for( i = 0; i <= trigger_count; i++) {
		/* if trigger on
		*/
		if ( trigger_cap[i].state == 1) {
			/* if trigger is not done storing packets
			*/
			if (trigger_cap[i].pcapdone == 0)  {
				/* flush it
				*/
				pcap_dump_flush(trigger_cap[i].pd_dump);
			}
		}  /* if trigger cap. state is on */ 
	} /*end for*/
} 
